# PY3 + PY2
from __future__ import unicode_literals
from __future__ import print_function

from moya.compat import text_type

import re
import unicodedata

_re_html = re.compile(r'<.*?>|\&.*?\;', re.UNICODE | re.DOTALL)


def escape(text):
    """Escape text for inclusion in html"""
    return text_type(text).replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;').replace("'", "&#39;")


def slugify(value):
    # Borrowed from Django
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^\w\s-]', '', value).strip().lower()
    return re.sub('[-\s]+', '-', value)


def textilize(s):
    """Remove markup from html"""
    s = s.replace("<p>", " ").replace('&nbsp;', ' ')
    return _re_html.sub("", s)


def summarize(s, max_size=100):
    s = textilize(s)

    if len(s) > max_size:
        if not s[max_size].isspace():
            words = s[:max_size].split()
            words.pop()
        else:
            words = s[:max_size].split()
        s = escape(' '.join(words)) + ' [&hellip;]'
    else:
        s = escape(s)
    return s


if __name__ == "__main__":
    print(escape("10 > 5 < 8 & foo"))

    print(summarize('Hello <a href="#">test</a> werwer'))

    # from moya.tools import timer, MultiReplace

    # replace = MultiReplace({'&': '&amp;', '<' : '&lt;', '>' : '&gt;', '"': '&quote;', "'": "&#39"})

    # text = '''Some "text" & 'stuff'<>'''

    # REPEAT = xrange(100000)


    # with timer('chained escape'):
    #     for _ in REPEAT:
    #         escape(text)

    # with timer('multi escape'):
    #     for _ in REPEAT:
    #         replace(text)



from __future__ import unicode_literals
from __future__ import absolute_import

from ..elements.elementbase import LogicElement, Attribute
from ..tags.context import DataSetter
from .. import logic
from .. import http
from ..tools import datetime_to_epoch, md5_hexdigest
from ..context.missing import is_missing
from .. import errors
from ..compat import text_type, PY2, py2bytes, urlencode, urlparse, parse_qs, urlunparse
from ..request import ReplaceRequest
from ..response import MoyaResponse

from fs.path import basename
from fs.errors import FSError

from datetime import datetime
import mimetypes
import time
import json

import logging
log = logging.getLogger('moya.runtime')


class ResponseTag(DataSetter):
    """Create a response object"""

    class Help:
        synopsis = "create a response object"

    class Meta:
        tag_name = "response"

    status = Attribute("Status code", type="httpstatus", required=False, default=200)
    content_type = Attribute("Content Type", default="text/html; charset=UTF-8")
    body = Attribute("Response body", type="expression", default=None)
    charset = Attribute("Character set", default="utf-8")
    headers = Attribute("Headers", type="dict", default=None)

    def get_value(self, context):
        let_map = self.get_let_map(context)
        (status,
         content_type,
         body,
         charset,
         headers) = self.get_parameters(context,
                                        'status',
                                        'content_type',
                                        'body',
                                        'charset',
                                        'headers')
        text = None
        if body is None:
            text = context.sub(self.text)
        response = MoyaResponse(status=status,
                                content_type=py2bytes(content_type),
                                body=body,
                                text=text,
                                charset=py2bytes(charset))
        for k, v in let_map.items():
            try:
                setattr(response, k, v)
            except:
                self.throw("bad-value.response-value",
                           "Can't set {} to {}".format(context.to_expr(k), context.to_expr(v)))
        if headers:
            response.headers.update(headers)
        return response


class Serve(LogicElement):
    """
    This tag serves a response object, which may be created with the [tag]response[/tag].

    """

    class Help:
        synopsis = "serve a response object"
        example = """
        <url route="/test/">
            <response status="im_a_teapot" dst="teapot_response">
                Short and Stout
            </response>
            <serve response="teapot_response" />
        </url>
        """

    response = Attribute("Response object to serve", type="expression", required=True)

    def logic(self, context):
        response = self.response(context)
        raise logic.EndLogic(response)


class ServeFile(LogicElement):
    """Serve a static file."""

    class Help:
        synopsis = "serve a file"
        example = """
        <serve-file fs="static" path="/images/logo.jpg />
        """

    path = Attribute("Path in filesystem", required=True)
    fsobj = Attribute("Filesystem object", type="Index")
    fs = Attribute("Filesystem name")
    ifexists = Attribute("Only serve a response if the file exists", type="boolean")
    name = Attribute("Name of the file", required=False, default=None)

    def logic(self, context):
        params = self.get_parameters(context)

        if params.fsobj is not None:
            fs = params.fsobj
        else:
            try:
                fs = self.archive.filesystems[params.fs]
            except KeyError:
                self.throw("serve.no-fs", "No filesystem called '{}'".format(params.fs))

        path = params.path

        if params.ifexists and not fs.isfile(path):
            return

        req = context.root["request"]
        res = MoyaResponse()

        mime_type, encoding = mimetypes.guess_type(basename(path))
        if mime_type is None:
            mime_type = b"application/octet-stream" if PY2 else "application/octet-stream"

        if not path or not fs.isfile(path):
            raise logic.EndLogic(http.RespondNotFound())

        serve_file = None
        try:
            file_size = fs.getsize(path)
            info = fs.getinfokeys(path, 'modified_time')
            serve_file = fs.open(path, 'rb')
        except FSError:
            if serve_file is not None:
                serve_file.close()
            raise logic.EndLogic(http.RespondNotFound())
        else:
            mtime = info.get('modified_time', None)
            if mtime is None:
                mtime = time.time()
            else:
                mtime = datetime_to_epoch(mtime)
            res.date = datetime.utcnow()
            res.content_type = py2bytes(mime_type)
            res.last_modified = mtime
            res.etag = "%i-%i-%s" % (mtime, file_size, md5_hexdigest(path))
            res.server = "Moya/1.0"
            if params.name:
                res.content_disposition = 'attachment; filename="{}"'.format(params.name)

            status304 = False
            if req.if_none_match and res.etag:
                status304 = res.etag in req.if_none_match
            elif req.if_modified_since and res.last_modified:
                status304 = res.last_modified <= req.if_modified_since
            if status304:
                res.status = 304
                serve_file.close()
            else:
                res.body_file = serve_file
            res.content_length = file_size
        raise logic.EndLogic(res)


class ServeJSON(LogicElement):
    """Serve an object encoded as JSON"""

    class Help:
        synopsis = """serve an object as JSON"""

    obj = Attribute("Object to build JSON from", type="index", required=False, default=None, missing=False)
    indent = Attribute("Indent to make JSON more readable", required=False, default=4)

    def logic(self, context):
        if self.has_parameter('obj'):
            obj = self.obj(context, context)
            json_obj = json.dumps(obj, indent=self.indent(context))
        else:
            json_obj = context.sub(self.text)
        response = MoyaResponse(content_type=b'application/json' if PY2 else 'application/json',
                                body=json_obj)
        raise logic.EndLogic(response)


class ServeXML(LogicElement):
    """Serve XML"""

    class Help:
        synopsis = """serve xml"""

    obj = Attribute("A string of XML, or an object that may be converted to XML", type="expression", required=True, missing=False)
    content_type = Attribute("Mime type", default="application/xml")

    def logic(self, context):
        params = self.get_parameters(context)
        mime_type = params.content_type
        xml = params.obj

        if hasattr(xml, '__xml__'):
            try:
                xml = xml.__xml__()
            except Exception as e:
                self.throw('serve-xml.fail', 'failed to covert {} to XML ({})'.format(context.to_expr(xml), e))

        if not isinstance(xml, bytes):
            xml = text_type(xml)
            xml_bytes = xml.encode('utf-8')
        else:
            xml_bytes = xml
        response = MoyaResponse(content_type=py2bytes(mime_type), body=xml_bytes)
        raise logic.EndLogic(response)


class NotFound(LogicElement):
    """Respond to the request with a '404 not found' response"""

    class Help:
        synopsis = "serve a '404 not found' response"

    def logic(self, context):
        raise logic.EndLogic(http.RespondNotFound())


class Forbidden(LogicElement):
    """Respond to the request with a '403 forbidden' response"""

    class Help:
        synopsis = "respond with a forbidden response"

    def logic(self, context):
        raise logic.EndLogic(http.RespondForbidden())


class RedirectTo(LogicElement):
    """Redirect to new location."""

    class Help:
        synopsis = "redirect to a new location"
        example = """
    <redirect-to url="http://www.moyaproject.com" />
    <redirect-to path="../newuser?result=success" />
    """

    url = Attribute("Destination URL", metavar="URL", required=False, default=None)
    path = Attribute("New path portion of the url, may be relative", metavar="PATH", required=False)
    code = Attribute("HTTP response code (use 301 for permanent redirect)", metavar="HTTPCODE", required=False, default=303, type="httpstatus")
    query = Attribute("Mapping expression to use as a query string", metavar="EXPRESSION", required=False, default=None, type="expression")
    fragment = Attribute("Fragment component in url")

    def logic(self, context):
        (url,
         path,
         query,
         fragment) = self.get_parameters(context,
                                         'url',
                                         'path',
                                         'query',
                                         'fragment')

        if url is not None:
            parsed_url = urlparse(url)
            url = urlunparse(parsed_url[0:3] + ('', '', ''))
            url_query = parsed_url.query
            query_components = parse_qs(url_query)
        else:
            query_components = {}

        request = context['.request']

        query_components.update(self.get_let_map(context))

        if query:
            query_components.update(query)
        if query_components:
            qs = urlencode(query_components, True)
            url += '?' + qs

        if url is not None:
            location = url
        elif path is not None:
            location = request.relative_url(path)
        if fragment:
            location = "{}#{}".format(location, fragment)
        self.new_location(context, location)

    def new_location(self, context, location):
        code = self.code(context)
        response = MoyaResponse(status=code)
        response.location = location
        raise logic.EndLogic(response)


class Redirect(LogicElement):
    """Redirect to a mounted URL"""

    class Help:
        synopsis = "redirect to a named URL"

    name = Attribute("URL name", required="y")
    _from = Attribute("Application", type="application", default=None)
    code = Attribute("HTTP response code (use 301 for permanent redirect)", metavar="HTTPCODE", required=False, default=303, type="integer")
    query = Attribute("Mapping expression to use as a query string", metavar="EXPRESSION", required=False, default=None, type="expression")
    fragment = Attribute("Fragment component in url")

    def logic(self, context):
        (urlname,
         query,
         fragment) = self.get_parameters(context,
                                         'name',
                                         'query',
                                         'fragment')
        app = self.get_app(context)

        if not app:
            raise errors.AppMissingError()

        app_name = app.name
        url_params = self.get_let_map(context)

        url_index = ".urls." + app_name + '.' + urlname

        with context.data_frame(url_params):
            try:
                url = context[url_index]
            except:
                self.throw('url.missing',
                           "No named URL called '{}' in application '{}'".format(urlname, app_name),
                           diagnosis="Check for typos in the url name. If you want to redirect to a URL in a different application, set the **from** attribute.",
                           name=urlname,
                           app=app)
            if is_missing(url):
                self.throw('url.missing',
                           "No named URL called '{}' in application '{}'".format(urlname, app_name),
                           diagnosis="Check for typos in the url name. If you want to redirect to a URL in a different application, set the **from** attribute.",
                           name=urlname,
                           app=app)

            url = text_type(context.get(url_index))

        if query:
            qs = urlencode(list(query.items()), True)
            url += '?' + qs

        location = url
        if fragment:
            location = "{}#{}".format(location)

        self.new_location(context, location)

    def new_location(self, context, location):
        code = self.code(context)
        response = MoyaResponse(status=code)
        response.location = location
        raise logic.EndLogic(response)


class Rewrite(Redirect):
    """
    This tag tells Moya to serve the content from a different named URL.

    Note, that unlike [tag]redirect[/tag], this does not involve an extra request.

    """

    class Help:
        synopsis = "serve response from a different named URL"

    code = None

    def new_location(self, context, location):
        url = urlparse(location)
        request = context['.request']
        new_request = request.copy()
        new_request.environ['QUERY_STRING'] = url.query
        new_request.environ['PATH_INFO'] = url.path
        log.debug('rewriting url to %s', location)
        raise logic.EndLogic(ReplaceRequest(new_request))


class RewriteTo(RedirectTo):
    """
    This tag tells Moya to serve the content from a different URL.

    Note, that unlike [tag]redirect-to[/tag], this does not involve an extra request.

    """

    code = None

    class Help:
        synopsis = """serve response from a different location"""

    def new_location(self, context, location):
        url = urlparse(location)
        request = context['.request']
        new_request = request.copy()
        new_request.environ['QUERY_STRING'] = url.query
        new_request.environ['PATH_INFO'] = url.path
        log.debug('rewriting url to %s', location)
        raise logic.EndLogic(ReplaceRequest(new_request))


class SetHeader(LogicElement):
    """
    Add additional headers to the outgoing response.

    """

    header = Attribute("Header name", required=True)
    value = Attribute("Header Value", required=False, default='')

    class Help:
        synopsis = "add additional headers"
        example = """
        <set-header header="moya-example">In your headerz</set-header>
        """

    def logic(self, context):
        header, value = self.get_parameters(context, 'header', 'value')
        if not self.has_parameter('value'):
            value = context.sub(self.text).strip()
        headers = context.set_new_call('.headers', dict)
        headers[header] = value



from ..markup import Markup, get_installed_markups
from ..elements.elementbase import Attribute
from ..tags.content import RenderBase
from ..tags.context import DataSetter
from ..compat import text_type

from textwrap import dedent


class _Markup(RenderBase):
    """Insert markup in to content"""

    class Help:
        synopsis = "insert markup in to content"

    type = Attribute("Markup type", required=False, default="bbcode", choices=get_installed_markups())
    source = Attribute("Markup source", required=False, default=None, type="expression")

    def logic(self, context):
        type = self.type(context)
        if not Markup.supports(type):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(type))
        options = self.get_let_map(context)
        source_text = self.source(context) if self.has_parameter('source') else self.text
        text = self.source(context) or Markup.sub(type, context, source_text, options)
        markup = Markup(text, type, options)
        context['.content'].add_renderable(self._tag_name, markup)


class MarkupTag(RenderBase):

    source = Attribute("Markup source", required=False, default=None, type="expression")
    dedent = Attribute("De-dent source (remove common leading whitespace)?", type="boolean", default=True)

    class Help:
        undocumented = True

    def logic(self, context):
        if not Markup.supports(self.markup):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(self.markup))
        options = self.get_let_map(context)
        source_text = self.source(context) if self.has_parameter('source') else self.text
        if not self.has_parameter('source') and self.dedent(context):
            source_text = dedent(source_text)
        sub_text = Markup.sub(self.markup, context, source_text, options)
        markup = Markup(sub_text, self.markup, options)
        context['.content'].add_renderable(self._tag_name, markup)


class ProcessMarkup(DataSetter):
    """Process a given markup in to text"""

    type = Attribute("Markup type", required=False, default="bbcode", choices=get_installed_markups())
    src = Attribute("Markup source", required=False, default=None, type="expression")
    dst = Attribute("Destination", type="reference", default=None)

    class Help:
        synopsis = "markup text"

    def logic(self, context):
        type = self.type(context)
        if not Markup.supports(type):
            self.throw('markup.unsupported', "markup type '{}' is not supported".format(type))
        options = self.get_let_map(context)
        source_text = self.src(context) if self.has_parameter('src') else self.text
        text = self.src(context) or Markup.sub(type, context, source_text, options)
        markup = Markup(text, type, options)
        result = markup.process()
        self.set_context(context, self.dst(context), result)


class BBCode(MarkupTag):
    """Add bbcode to content"""
    markup = "bbcode"

    class Help:
        synopsis = "add bbcode to content"


class Markdown(MarkupTag):
    """Add markdown to content"""
    markup = "markdown"

    class Help:
        synopsis = "add markdown to content"



from __future__ import unicode_literals


from ..tags.context import DataSetter
from .. import errors


class ElementContainer(object):

    def __init__(self, archive):
        self.archive = archive

    def __getitem__(self, elementid):
        return self.archive.get_element(elementid)

    def __contains__(self, elementid):
        try:
            self[elementid]
        except errors.ElementNotFoundError:
            return False
        else:
            return True


class Inspect(DataSetter):
    """
    Get an object which contains all the elements in the project.
    """

    class Help:
        synopsis = "inspect elements in the project"

    def get_value(self, context):
        return ElementContainer(self.archive)



from __future__ import unicode_literals
from __future__ import print_function

from ..elements.registry import Meta
from ..elements import attributetypes
from ..elements.elementbase import ElementBase, Attribute, LogicElement
from ..logic import DeferNodeContents
from ..tools import textual_list, nearest_word
from ..context.missing import is_missing
from ..context.tools import to_expression
from ..compat import text_type, py2bytes
from .. import errors
from .. import namespaces


class WidgetBase(LogicElement):
    _container = False

    class Meta:
        text_nodes = "text"
        is_call = True

    def __init__(self, *args, **kwargs):
        self.init_args = (args, kwargs)
        super(WidgetBase, self).__init__(*args, **kwargs)

    def get_widget_app(self, context):
        app = None
        if self.has_parameter('from'):
            app = self.get_parameter(context, 'from') or None
        if app is None:
            try:
                app = self.archive.get_app_from_lib(self._lib_long_name)
            except:
                app = context.get('.app', None)
        if app is None:
            self.throw("widget.ambiguous-app",
                       "unable to detect app for this widget (you may need to supply the 'from' attribute)")
        return app

    def logic(self, context):
        widget_frame = context.current_frame
        content = context['.content']
        if is_missing(content):
            self.throw('widget.content-missing',
                       "widgets must be called in a content definition")

        td = self.get_all_parameters(context)
        td.pop('_if', None)
        let_map = self.get_let_map(context)
        widget_app = self.get_widget_app(context)

        def on_yield(context, app, content, element, data):
            context.push_frame(widget_frame)
            try:
                with self.defer(context, app=app):
                    with context.data_scope(data):
                        with content.node():
                            yield DeferNodeContents(element)
            finally:
                context.pop_frame()

        app = self.get_app(context)
        custom_template = app.resolve_template(self.template(context))
        template_node = content.add_template(self._tag_name,
                                             custom_template or widget_app.resolve_template(self._template),
                                             app=widget_app)

        self.push_call(context, td, widget_app)
        if self._let_dst:
            context[self._let_dst] = let_map

        yield_stack = context.set_new_call('._yield_stack', list)
        yield_stack.append(lambda c, data: on_yield(c, app, content, self, data))
        try:
            if self.widget_element.has_children:
                yield DeferNodeContents(self.widget_element)
        finally:
            yield_stack.pop()
            scope = context.obj
            self.pop_call(context)

        if self._container and self.has_children:
            with content.node():
                yield DeferNodeContents(self)

        if '_return' in scope:
            scope = scope['_return']
            if hasattr(scope, 'get_return_value'):
                scope = scope.get_return_value()
            if not scope:
                scope = {}
            if not hasattr(scope, 'items'):
                self.throw('widget.return-no-dict',
                           'the return value from a widget must be a dict, or None (not {})'.format(context.to_expr(scope)))

        template_node.update_template_data(scope)


class WidgetBaseContainer(WidgetBase):
    _container = True


class WidgetYield(LogicElement):
    """Yield to a widget block"""

    class Help:
        synopsis = "yield to code block in a widget"

    obj = Attribute("Data to yield", type="expression", required=False, default=None)

    def logic(self, context):
        yield_call = context.get_stack_top('yield')
        yield_data = self.obj(context) or {}
        yield_data.update(self.get_let_map(context))
        if callable(yield_call):
            for node in yield_call(context, yield_data.copy()):
                yield node


class Widget(ElementBase):
    """Create a widget"""

    class Help:
        synopsis = "create a re-uasable widget"

    ns = Attribute("XML Namespace", required=False)
    name = Attribute("Tag name", required=True)
    template = Attribute("Template", type="template", required=False, default=None)
    let = Attribute("Let destination", required=False, default=None)
    container = Attribute("Is this widget a container?", type="boolean", default=True, required=False)
    synopsis = Attribute("Short description of the widget")
    undocumented = Attribute("Set to yes to disabled documentation for this tag", type="boolean", default=False)

    def finalize(self, context):
        params = self.get_parameters(context)

        attributes = {}
        attributes['template'] = Attribute('Override widget template', name="template", required=False, default=None)
        for signature in self.children("signature"):
            for attribute_tag in signature.children("attribute"):
                param_map = attribute_tag.get_all_parameters(context)
                description = attribute_tag.doc
                name = param_map.pop('name')
                attribute = Attribute(description, name=name, evaldefault=True, **param_map)
                attributes[attribute.name] = attribute

        #if 'from' not in attributes:
        attributes['from'] = Attribute("Application", name="from", type="application", map_to="from", default=None)

        doc = None
        for doc_tag in self.children("doc"):
            doc = doc_tag.text.strip()
        meta = Meta()
        meta.is_call = True

        class Help(object):
            synopsis = params.synopsis
            undocumented = params.undocumented

        cls_dict = dict(__doc__=text_type(doc or ''), Meta=meta, Help=Help)
        if self.source_line:
            definition = "%s (line %s)" % (self._location, self.source_line)
        else:
            definition = self._location

        cls_dict['_definition'] = definition
        cls_dict['_template'] = params.template
        cls_dict['_let_dst'] = params.let
        cls_dict['xmlns'] = params.ns or self.lib.namespace or namespaces.default
        cls_dict.update(('_attribute_' + k, v)
                        for k, v in attributes.items())
        cls_dict['Meta'] = WidgetBase.Meta
        cls_dict['_registry'] = self.archive.registry

        if params.container:
            bases = (WidgetBaseContainer,)
        else:
            bases = (WidgetBase,)

        tag_class = type(py2bytes(params.name),
                         bases,
                         cls_dict)
        tag_class.widget_element = self
        tag_class.libname = None
        tag_class._definition = definition
        tag_class._lib_long_name = context.get('._lib_long_name', None)


class AttributeTag(ElementBase):
    """Defines an attribute in a [tag]tag[/tag], [tag]data-tag[/tag] or [tag]widget[/tag]."""

    class Help:
        synopsis = "define an attribute in a custom tag"
        example = """
        <datatag name="module">
            <doc>Define a top level admin module</doc>
            <signature>
                <attribute name="slug" required="yes" />
                <attribute name="title" required="yes" />
                <attribute name="description" required="yes" />
                <attribute name="content" type="elementref" required="no" />
            </signature>
        </datatag>

        """

    _element_class = "widget"
    preserve_attributes = ['doc']

    class Meta:
        logic_skip = True
        tag_name = "attribute"
    name = Attribute("Name of the attribute", required=True)
    type = Attribute("Type of the attribute", required=False, choices=attributetypes.valid_types)
    required = Attribute("Required", type="boolean", required=False, default=False)
    default = Attribute("Default", required=False, default=None)
    metavar = Attribute("Metavar (identifier used in documentation)", required=False)
    missing = Attribute("Are missing values allowed?", type="boolean", default=True, required=False)
    choices = Attribute("Valid values for this attribute", type="commalist", default=None, required=False)

    def post_build(self, context):
        self.doc = context.sub(self.text.strip())


class ArgumentTag(ElementBase):
    """
    Defines an argument to a macro.

    The text of this tag should document the purpose of the argument.

    """

    class Help:
        synopsis = "define an argument to a macro"
        example = """
        <macro docname="average">
            <signature>
                <argument name="numbers" required="yes" check="len:numbers gt 0">
                    A list (or other sequence) of numbers
                </argument>
            </signature>
            <return value="sum:numbers / len:numbers" />
        </macro>

        """

    class Meta:
        logic_skip = True
        tag_name = "argument"

    name = Attribute("Name of the attribute", required=True)
    required = Attribute("Is this argument required?", type="boolean", default=False)
    check = Attribute("A boolean expression that the attribute must satisfy", type="function", default=None)

    def post_build(self, context):
        self.doc = context.sub(self.text.strip())


class ArgumentValidator(object):
    """Checks arguments to a macro call."""

    def __init__(self, context, element):
        self.doc = []
        self.required = []
        self.checks = []
        self.arg_names = set()
        for arg in element.children():
            if arg._element_type != (namespaces.default, 'argument'):
                raise errors.ElementError("must contain <argument> tags", element=element)
            name, required, check = arg.get_parameters(context, 'name', 'required', 'check')
            self.arg_names.add(name)
            if required:
                self.required.append(name)
            if check is not None:
                self.checks.append((name, check))
            self.doc.append({"name": name, "required": required, "check": check})
        self.required_set = frozenset(self.required)

    def __repr__(self):
        if self.arg_names:
            return "<validator {}>".format(textual_list(self.arg_names))
        else:
            return "<validator>"

    def validate(self, context, element, arg_map):
        if not self.arg_names.issuperset(arg_map.keys()):
            for k in arg_map:
                if k not in self.arg_names:
                    nearest = nearest_word(k, self.arg_names)
                    if nearest is not None:
                        diagnosis = "Did you mean '{}'?".format(nearest)
                    else:
                        diagnosis = "Valid arguments to this macro are {}.".format(textual_list(sorted(self.arg_names), 'and'))
                    element.throw('argument-error.unknown-argument',
                                  "no argument called '{name}' in {element}'".format(name=k, element=element),
                                  diagnosis=diagnosis)
        if not self.required_set.issubset(arg_map.keys()):
            for name in self.required:
                if name not in arg_map:
                    element.throw('argument-error.required',
                                  "argument '{}' is required in {}".format(name, element),
                                  diagnosis='''You can pass a value for '{name}' with let:{name}="<VALUE>" '''.format(name=name, element=element))
        for name, check in self.checks:
            try:
                result = check.call(context, arg_map)
            except Exception as e:
                element.throw('argument-error.check-error',
                              "check for argument '{}' in {} failed with exception: {}".format(name, element, e),
                              diagnosis="An exception was thrown when evaluating the expression '{}'.\n\n"
                              "This could indicate a programming error in the macro or Moya.".format(check.expression))
            if not result:
                diagnosis_msg = "{value} is not a valid value for argument '{name}'. Check the calling logic is correct."
                element.throw('argument-error.check-failed',
                              '''argument '{}' failed check "{}" in {}'''.format(name, check.expression, element),
                              diagnosis=diagnosis_msg.format(name=name,
                                                             value=to_expression(context, arg_map[name]),
                                                             element=element))


class Signature(ElementBase):
    """
    Begins a list of attributes and arguments for a [tag]tag[/tag], [tag]data-tag[/tag], [tag]macro[/tag] or [tag]command[/tag].

    In the case of tags, the signature should contain [tag]attribute[/tag] tags.
    Macros expect [tag]argument[/tag] tags.
    For a command, the signature should contain [tag]arg[/tag] and [tag]option[/tag] tags.

    """
    _element_class = "widget"

    class Help:
        synopsis = "define the attributes / arguments to a tag / macro"
        example = """
        <tag name="fib">
            <doc>Calculate the fibonacci sequence</doc>
            <signature>
                <attribute name="count" type="integer" />
            </signature>
            <let fib="[0, 1]"/>
            <repeat times="count - 2">
                <append value="fib[-1] + fib[-2]" dst="fib" />
            </repeat>
            <return value="fib" />
        </tag>

        """

    class Meta:
        logic_skip = True

    def finalize(self, context):
        if self.parent._element_type == (namespaces.default, 'macro'):
            self.validator = ArgumentValidator(context, self)


class Doc(ElementBase):
    """
    Write documentation for a widget or custom tag.

    """
    _element_class = "doc"

    class Meta:
        logic_skip = True

    class Help:
        synopsis = "document a tag"
        example = """
        <widget name="post" template="post.html">
            <doc>Renders a single post</doc>
            <!-- widget code -->
        </widget>

        """



from __future__ import unicode_literals
from __future__ import print_function

import moya
from moya.compat import text_type, PY2, py2bytes

import hashlib
import random
import time
try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    using_sysrandom = False


from passlib.context import CryptContext

if PY2:
    moya_pwd_context = CryptContext(
        schemes=[b"bcrypt", b"sha512_crypt", b"pbkdf2_sha512"],
        default=b"pbkdf2_sha512",
        all__vary_rounds=0.1,
    )
else:
    moya_pwd_context = CryptContext(
        schemes=["bcrypt", "sha512_crypt", "pbkdf2_sha512"],
        default="pbkdf2_sha512",
        all__vary_rounds=0.1,
    )

UNUSABLE_PASSWORD = "!"


def get_random_string(length=12,
                      allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
                      secret=''):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            hashlib.sha256(
                "%s%s%s" % (
                    random.getstate(),
                    time.time(),
                    secret)
            ).digest())
    return ''.join([random.choice(allowed_chars) for i in range(length)])


@moya.expose.macro("makesessionkey")
def makesessionkey(app):
    return get_random_string(secret=app.settings.get('secret', ''))


@moya.expose.macro("hashpassword")
def hashpassword(app, password):
    rounds = app.settings.get_int('rounds', 10000)
    scheme = text_type(app.settings.get('scheme', 'pbkdf2_sha512'))
    try:
        password_hash = moya_pwd_context.encrypt(password,
                                                 scheme=py2bytes(scheme),
                                                 rounds=rounds)
    except Exception as e:
        app.throw('moya.auth.password-hash-fail', text_type(e))
    return password_hash


@moya.expose.macro("checkpassword")
def checkpassword(app, password, stored_password):
    if not password or password == UNUSABLE_PASSWORD:
        return False
    try:
        passed = moya_pwd_context.verify(password, stored_password)
    except Exception:
        app.log.exception('error when verifying password')
        passed = False
    return passed



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from ..elements.elementbase import (LogicElement,
                                    ReturnContainer,
                                    Attribute)
from ..context import Expression
from ..context.expressiontime import ExpressionDateTime, TimeSpan
from ..context import dataindex
from ..context.errors import ContextKeyError
from ..containers import OrderedDict
from ..progress import Progress
from ..tools import make_cache_key
from .. import namespaces
from ..logic import (DeferNodeContents,
                     SkipNext,
                     Unwind,
                     EndLogic,
                     BreakLoop,
                     ContinueLoop,
                     MoyaException)
from ..console import style as console_style
from ..render import HTML
from ..compat import zip_longest, raw_input, text_type, string, xrange

import json
import weakref
import getpass
import time
from collections import defaultdict
from datetime import datetime
from random import choice

from time import sleep
from textwrap import dedent

import sys
import logging
log = logging.getLogger('moya.runtime')


try:
    import readline
except ImportError:
    pass


class ContextElementBase(LogicElement):
    pass


class If(ContextElementBase):
    """
    Conditional [i]IF[/i] tag, executes the enclosing block if a condition evaluates to [i]true[/i]. If the condition evaluates to [i]false[/i], the enclosing block is skipped. May be followed by [tag]else[/tag] or [tag]elif[/tag].

    """
    class Help:
        synopsis = """execute block if a condition is true"""
        example = """
        <if test="coffee==0">
            <echo>Get more coffee</echo>
        </if>"""

    test = Attribute("Test expression", required=True, metavar="CONDITION", type="expression")

    def logic(self, context):
        if self.test(context):
            yield DeferNodeContents(self)
            yield SkipNext((self.xmlns, "else"), (self.xmlns, "elif"))


class IfPost(ContextElementBase):
    """
    Executes the enclosing block if the current request is a POST request.

    Basically a shorthand for:

    [code xml]<if test=".request.method=='POST'">[/code]

    May be followed by [tag]else[/tag] or [tag]elif[/tag].

    """
    class Help:
        synopsis = """execute a block if this is a POST request"""
        example = """\
<if-post>
    <!-- validate form -->
</if-post>
<else>
    <!-- not a post request -->
</else>"""

    def logic(self, context):
        if context.get('.request.method', None) == 'POST':
            yield DeferNodeContents(self)
            yield SkipNext((self.xmlns, "else"), (self.xmlns, "elif"))


class IfGet(ContextElementBase):
    """
    Executes the enclosing block if the current request is a GET request.

    Basically a shorthand for:

    [code xml]<if test=".request.method=='GET'">[/code]

    May be followed by [tag]else[/tag] or [tag]elif[/tag].

    """
    class Help:
        synopsis = """exectute a block if this is a GET request"""

    def logic(self, context):
        if context.get('.request.method', None) == 'GET':
            yield DeferNodeContents(self)
            yield SkipNext((self.xmlns, "else"), (self.xmlns, "elif"))


class Else(ContextElementBase):
    """Execute the enclosed block if a previous [tag]if[/tag] statement is false."""

    class Help:
        synopsis = """execute a block if the previous <if> statement is false"""
        example = """
<if test="coffee==0">
    <echo>Get more coffee</echo>
</if>
<else>
    <echo>Have a cup of coffee</echo>
</else>

"""

    def logic(self, context):
        yield DeferNodeContents(self)


class Elif(ContextElementBase):
    """Executes the enclosed block if a previous [tag]if[/tag] (or [tag]elif[/tag]) statement is false, and another condition is true."""
    class Help:
        synopsis = """an <else> with a condition"""
        example = """\
<if test="hobbits == 0">
    <echo>There are no hobbits</echo>
</if>
<elif test="hobbits == 1">
    <echo>There is one hobbit</echo>
</elif>
<else>
    <echo>There are many hobbits</echo>
</else>
"""
    test = Attribute("Test expression", required=True, metavar="CONDITION", type="expression")

    def logic(self, context):
        if self.test(context):
            yield DeferNodeContents(self)
            yield SkipNext((self.xmlns, "else"), (self.xmlns, "elif"))


class Try(ContextElementBase):
    """Executes the enclosed block as part of a [tag]try[/tag][tag]catch[/tag] structure. """
    class Meta:
        is_try = True

    class Help:
        synopsis = """detect exceptions within a block"""
        example = """
<try>
    <echo>${1/0}</echo>  <!-- will throw a math exception -->
</try>
<catch exception="*">
    <echo>The try block threw an exception</echo>
</catch>
"""

    def logic(self, context):
        yield DeferNodeContents(self)


class ExceptionProxy(object):
    def __init__(self, msg, type, info):
        self.msg = msg
        self.type = type
        self.info = info

    def __repr__(self):
        return "<exception '{}'>".format(self.msg)


class Catch(ContextElementBase):
    """Catches any [link moyacode#exceptions]exceptions[/link] from the previous block."""

    class Help:
        synopsis = """catch exceptions from the previous block"""
        example = """
<try>
    <echo>${1/0}</echo>  <!-- will throw a math exception -->
</try>
<catch exception="math.divistion-error">
    <echo>The try block threw an exception</echo>
</catch>

<!-- A <try> block is only necessary if there are multiple statements you want to catch exceptions for -->
<echo>${1/0}<echo>
<catch exception="math.division-error">
    <echo>You can't divide by zero!</echo>
</catch>

        """

    exception = Attribute("Type of exception to catch", type="commalist", default="*", evaldefault=True)
    dst = Attribute("Destination to store exception object", type="reference")

    class Meta:
        logic_skip = True

    @classmethod
    def compare_exception(cls, type, catch_types):
        type_tokens = type.split('.')
        for catch_type in catch_types:
            for c, e in zip_longest(catch_type.split('.'),
                                    type_tokens,
                                    fillvalue=None):
                if c == '*':
                    return True
                if c != e:
                    break
            else:
                return True
        return False

    def check_exception_type(self, context, type):
        catch_type = self.exception(context)
        return self.compare_exception(type, catch_type)

    def set_exception(self, context, exception):
        dst = self.dst(context)
        if dst is not None:
            context[dst] = ExceptionProxy(exception.msg,
                                          exception.type,
                                          exception.info)


class Attempt(ContextElementBase):
    """Repeat a block of code if an exception is thrown. This is useful in situations where you are creating a database object
    with a random token, and you want to repeat the code if the random token happens to exist. The [c]wait[/c] attribute can set a timespan
    to wait between attempts, which may be appropriate for connecting to servers, for example.

    If an exception occurs, after [c]times[/c] attempts, then it will be thrown normally.

    [aside]Always set [c]times[/c] to a reasonable value. Programming errors could result in Moya getting stuck in a pointless loop.[/aside]

    """

    class Help:
        synopsis = """repeat a block until it is successful"""
        example = """
        <attempt times="10" catch="db.*">
            <db:create model="#ValidateEmail" let:user="user" dst="validate_email" />
        </attempt>

        """

    times = Attribute("Maximum number of times to repeat attempt block", type="integer", required=True)
    catch = Attribute("Type of exception to catch", type="commalist", default="*", evaldefault=True)
    wait = Attribute("Time to wait between syncs (default doesn't wait)", type="timespan", default=None, required=False)

    class Meta:
        trap_exceptions = True

    def on_exception(self, context, exception):
        stack = context.set_new_call('._attempt_stack', list)
        try:
            top = stack[-1]
        except IndexError:
            return None
        if not Catch.compare_exception(exception.type, top['type']):
            return None
        top['times'] -= 1
        if top['times'] > 0:
            if top['wait']:
                time.sleep(top['wait'])
            return DeferNodeContents(self)

    def logic(self, context):
        attempt_times = self.times(context)
        exception_type = self.catch(context)
        wait = self.wait(context)
        stack = context.set_new_call('._attempt_stack', list)
        stack.append({'times': attempt_times,
                      'type': exception_type,
                      'wait': wait})
        try:
            yield DeferNodeContents(self)
        except:
            stack.pop()


class With(ContextElementBase):
    """Merge values with the current scope in the enclosed block. The values in the let map will persist only in the enclosed block."""

    class Help:
        synopsis = """create a new data scope"""
        example = """
<with let:hobbit="Sam">
    <echo>${hobbit} is my favorite hobbit.
</with>
<!-- 'hobbit' will no longer exist at this point -->
<echo if="missing:hobbit">Where is the hobbit?</echo>

"""

    def logic(self, context):
        let_map = self.get_let_map(context)
        with context.data_scope(let_map):
            yield DeferNodeContents(self)


class DataSetterBase(ContextElementBase):
    default = None
    dst = Attribute("Destination", type="reference", default=None)

    class Help:
        undocumented = True

    def get_default(self):
        return self.default

    def set_context(self, context, dst, value):
        if dst is None:
            obj = context.obj
            append = getattr(obj, 'append', None)
            if append is not None:
                append(value)
                return text_type(len(obj) - 1)
            else:
                return dst
        else:
            context[dst] = value
            return dst

    def logic(self, context):
        self.set_context(context,
                         self.dst(context),
                         self.get_value(context))

    def get_value(self, context):
        if self.text:
            return self.process_value(context, self.text)
        return self.default

    def process_value(self, context, value):
        return value


class DataSetter(DataSetterBase):
    default = None
    dst = Attribute("Destination", type="reference", default=None)
    value = Attribute("Value", type="expression", default=None)

    class Help:
        undocumented = True


class Var(DataSetter):
    """Set a variable"""

    class Help:
        synopsis = """create a value"""
        example = """
<var dst="count">10</var>
"""

    def logic(self, context):
        dst, value = self.get_parameters(context, 'dst', 'value')
        if not self.has_parameter('value'):
            value = context.eval(context.sub(self.text))
        self.set_context(context, dst, value)


class SetItem(LogicElement):
    """Set a value in a container"""

    class Help:
        synopsis = """set an indexed value on a collection"""
        example = """
        <dict dst="moya" />
        <set-item src="moya" index="'crew'" value="['john', 'rygel'] />
        """

    src = Attribute("collection object", type="expression", required=True)
    index = Attribute("index to set", type="expression", required=True)
    value = Attribute("value to set", type="expression", required=True)

    # TODO: handle errors
    def logic(self, context):
        src, index, value = self.get_parameters(context, 'src', 'index', 'value')
        src[index] = value


class Let(DataSetter):
    """Sets multiple variable from expressions."""

    class Help:
        synopsis = """create variables from expressions"""
        example = """
<let count="10" hobbit="'Sam'" hobbits="count * hobbit" />
        """

    class Meta:
        all_attributes = True

    _reserved_attribute_names = ['if']
    preserve_attributes = ['expressions']

    def post_build(self, context, _parse=dataindex.parse, _Expression=Expression):
        self.expressions = []
        append = self.expressions.append
        for k, v in self._attributes.items():
            if k not in self._reserved_attribute_names:
                append((_parse(k), _Expression(v).eval))

    def logic(self, context):
        try:
            context.set_multiple((k, v(context))
                                 for k, v in self.expressions)
        except ContextKeyError as e:
            self.throw('let.fail',
                       text_type(e))


class LetStr(DataSetter):
    """Like [tag]let[/tag] but sets variables as strings. Setting strings with [tag]let[/tag] can look a little clumsy because of the requirement to escape the text twice. [tag]let-str[/tag] treats attributes as strings and not expressions."""

    class Help:
        synopsis = """create string variables"""
        example = """
<let-str hobbit="Sam" dwarf="Durin" />
"""

    class Meta:
        all_attributes = True

    _reserved_attribute_names = ['if']
    preserve_attributes = ['expressions']

    def post_build(self, context):
        self.expressions = []
        append = self.expressions.append
        for k, v in self._attributes.items():
            if k not in self._reserved_attribute_names:
                append((dataindex.parse(k), v))

    def logic(self, context):
        sub = context.sub
        for k, v in self.expressions:
            context[k] = sub(v)


class Int(DataSetter):
    """Set an integer value."""
    default = 0

    class Help:
        synopsis = """create an integer"""
        example = """
<int dst="count" value="10"/>
<int dst="position" /> <!-- defaults to 0 -->

"""

    def process_value(self, context, value):
        return int(value)

    def logic(self, context):
        dst, value = self.get_parameters(context, 'dst', 'value')
        if not self.has_parameter('value'):
            value = context.sub(self.text) or 0
        self.set_context(context,
                         dst,
                         int(value))


class Float(DataSetter):
    """Set a float value"""
    default = 0.0

    class Help:
        synopsis = """create a floating point value"""
        example = """
    <int dst="count" value="10"/>
    <int dst="position" /> <!-- defaults to 0.0 -->

    """

    def logic(self, context):
        dst, value = self.get_parameters(context, 'dst', 'value')
        if not self.has_parameter('value'):
            value = context.sub(self.text) or 0.0
        self.set_context(context,
                         dst,
                         float(value))


class Str(DataSetter):
    """Sets a string value to text contexts."""

    class Help:
        synopsis = """create a string"""
        example = """
<str dst="hobbit">Sam</str>
<str dst="text"/> <!-- defaults to empty string -->
"""

    default = ''
    value = Attribute("Value", type="expression", default=None)

    def get_value(self, context):
        if self.has_parameter('value'):
            return self.value(context)
        return context.sub(self.text)


class Dedent(DataSetter):
    """Removes common whitespace from the start of lines."""

    class Help:
        synopsis = """remove common whitespace from a string"""
        example = """
        <dedent dst="text">
            This text will have the initial whitespace removed
                This will start with only 4 spaces
            Back to column 0
        </dedent>
        """

    def get_value(self, context):
        text = context.sub(self.value(context) or self.text)
        dedent_text = dedent(text)
        return dedent_text


class HTMLTag(DataSetter):
    """Sets a variable to HTML."""

    class Help:
        synopsis = "create raw html"
        example = """
        <html dst="my_html"><![CDATA[
           <h1>Hello, World!</h1>
        ]]></html>
        """

    class Meta:
        tag_name = "html"

    def get_value(self, context):
        if self.has_parameter('value'):
            text = self.value(context)
        else:
            text = context.sub(self.text)
        return HTML(text)


# class GetLet(DataSetter):
#     """Retrieve the [i]let map[/i] of the enclosing tag."""

#     def get_value(self, context):
#         let_map = self.parent.get_let_map(context)
#         return let_map


# class RenderBBCode(DataSetter):
#     default = ''

#     def get_value(self, context):
#         from postmarkup import render_bbcode
#         return HTML(render_bbcode(context.sub(self.text)))


class _TimeSpan(DataSetter):
    """Create a datetime object"""

    class Help:
        synopsis = "create a timespan object"
        example = """
        <time-span dst="span" value="1d"/>
        """

    value = Attribute("Value", type="text", default="")

    def get_value(self, context):
        text = self.value(context) or context.sub(self.text)
        if not text:
            self.throw('time-span.invalid', "no timespan value specified")
        try:
            return TimeSpan(text)
        except ValueError as e:
            self.throw('time-span.invalid', text_type(e))


class Datetime(DataSetter):
    """
    Create a Moya [i]datetime[/i] object.

    Datetime objects have the following properties:

    [definitions]
        [define year]The year (integer)[/define]
        [define month]The month (integer) (1 is January)[/define]
        [define day]The day (integer)[/define]
        [define hour]The hours in the time[/define]
        [define minute]The number of minutes in the time[/define]
        [define second]The number of seconds in the time[/define]
        [define microseconds]The number of microseconds in the time[/define]
        [define isoformat]The date/time in [url http://en.wikipedia.org/wiki/ISO_8601]ISO 8601[/url] format[/define]
        [define next_day]The date/time + 24 hours[/define]
        [define previous_day]The date/time - 24 hours[/define]
        [define year_start]January 1st at the beginning of the year[/define]
        [define day_start]The beginning of the day (midnight)[/define]
        [define next_year]January 1st of the next year[/define]
        [define next_month]First day of the next month[/define]
        [define next_day]The start of the next day[/define]
        [define previous_day]The start of the previous day[/define]
        [define previous_year]Jan 1st of the previous year[/define]
        [define previous_monthy]First date of the previous month[/define]
        [define days_in_month]Number of days in the current month[/define]
        [define epoch]Number of seconds since Jan 1st 1970[/define]
        [define html5_datatime]Date/Time in HTML5 format[/define]
        [define html5_data]Date in HTML5 format[/define]
        [define html_time]Time in HTML5 format[/define]
        [define utc]Date/Time in UTC timezone[/define]
        [define local]DateTime in local timezone (references [c].tz[/c])[/define]
    [/definitions]

    """

    class Help:
        synopsis = """create a date/time object"""
        example = """
        <datetime year="1974" month="7" day="5" dst="birthday" />
        <echo>Your birthday is ${localize:birthday}</echo>
        """

    year = Attribute("Year", type="integer", required=True)
    month = Attribute("Month", type="integer", required=False, default=1)
    day = Attribute("Month", type="integer", required=False, default=1)
    hour = Attribute("Hour", type="integer", required=False, default=0)
    minute = Attribute("Minute", type="integer", required=False, default=0)
    second = Attribute("Second", type="integer", required=False, default=0)

    def get_value(self, context):
        try:
            return ExpressionDateTime(*self.get_parameters(context,
                                                           "year",
                                                           "month",
                                                           "day",
                                                           "hour",
                                                           "minute",
                                                           "second"))
        except Exception as e:
            self.throw("datetime.invalid", text_type(e))


class Bool(DataSetter):
    """Create a boolean (True or False value)."""

    class Help:
        synopsis = """create a boolean"""
        example = """
        <bool dst="bool"/> <!-- False -->
        <bool dst="bool">yes</bool> <!-- True -->
        <bool dst="bool">no</bool> <!-- False -->
        <bool dst="bool" value="1"/> <!--True -->
        <bool dst="bool" value="0"/> <!-- False -->
        """

    default = False

    def logic(self, context):
        if self.text:
            value = self.text.lower() in ('yes', 'true')
        else:
            value = self.value(context)
        self.set_context(context,
                         self.dst(context),
                         bool(value))


class True_(DataSetter):
    """Creates a boolean with a value of True."""

    class Help:
        synopsis = """create a True boolean"""
        example = """
        <true dst="bool"/> <!-- Always True -->
        """

    default = True

    def get_value(self, context):
        return True


class False_(DataSetter):
    """Creates a boolean with a value of False."""

    class Help:
        synopsis = """create a False boolean"""
        example = """
        <false dst="bool"/> <!-- Always False -->
        """
    default = False

    def get_value(self, context):
        return False


class None_(DataSetter):
    """Create a None value."""

    class Help:
        synopsis = """Create a value of None"""
        example = """
        <none dst="hobbits"/>
        <echo>Hobbits in England: ${hobbits}</echo>
        """

    default = None

    def get_value(self, context):
        return None


class Now(DataSetter):
    """Create a datetime objects for the current time. The datetime is in UTC, use the [c]localize:[/c] modifier to display a time in the user's current timezone."""

    class Help:
        synopsis = """get the current date/time"""
        example = """
        <now dst="now" />
        <echo>The time is ${localize:now}</now>
        """

    def logic(self, context):
        self.set_context(context,
                         self.dst(context),
                         datetime.utcnow())


class List(DataSetter):
    """Creates a list. Any data items in the enclosed block are added to the list."""

    class Help:
        synopsis = """create a list"""
        example = """
        <list dst="hobbits">
            <str>Sam</str>
            <str>Bilbo</str>
            <str>Frodo</str>
        </list>
        <echo>${commalist:hobbits}</echo>
        <list dst="empty"/>
        """

    def logic(self, context):
        dst = self.set_context(context, self.dst(context), [])
        context.push_scope(dst)
        try:
            yield DeferNodeContents(self)
        finally:
            context.pop_scope()


class Sum(DataSetter):
    """Sums a sequence of values together."""

    class Help:
        synopsis = """sums a sequence of values"""
        example = """
        <sum dst="moluae">
            <int>2</int>
            <int>40</int>
        </sum>
        <echo>${moluae}</echo>
        """

    def logic(self, context):
        container = []
        list_index = self.set_context(context, self.dst(context), container)
        context.push_scope(list_index)
        try:
            yield DeferNodeContents(self)
        finally:
            context.pop_scope()
        self.set_context(context, self.dst(context), sum(container))


class AppendableSet(set):
    """A set with an 'append' method that is an alias for 'add'"""
    def append(self, value):
        return self.add(value)


class Set(DataSetter):
    """Create a [i]set[/i] object. A set is a container, where each item may only appears once. Any data items in the enclosed block are added to the set."""

    class Help:
        synopsis = """create a set object"""
        example = """
        <list dst="hobbits">
            <str>Sam</str>
            <str>Bilbo</str>
            <str>Frodo</str>
            <str>Bilbo</str>
            <str>Sam</str>
        </list>
        <!-- Will display 3, because duplicates are removed -->
        <echo>There are ${len:hobbits} in the set.</echo>

        """

    def logic(self, context):
        dst = self.set_context(context, self.dst(context), AppendableSet())
        context.push_scope(dst)
        try:
            yield DeferNodeContents(self)
        finally:
            context.pop_scope()


class Dict(DataSetter):
    """Create a [i]dictionary[/i] object. A dictionary maps [i]keys[/i] on to [i]values[/i]. The keys and values are defined in the enclosing block, or via the [i]let map[/i]."""

    class Help:
        synopsis = """create a dict object"""
        example = """
        <dict dst="species">
            <str dst="Bibo">Hobbit</dst>
            <str dst="Gandalph">Wizard</dst>
        </dict>
        <!-- Alternatively -->
        <dict dst="species" let:Bilbo="Hobbit" let:Gandalph="Wizard" />
        <echo>Bilbo is a ${species['Bilbo']}</echo>

        """

    default = Attribute("Default object to return if the key isn't in the dict", required=False, default=None, choices=['dict', 'list', 'int', 'float'])
    sequence = Attribute("Optional sequence of key / value pairs to initialize the dict", type="expression", required=False)

    _default_types = {'dict': OrderedDict,
                      'list': list,
                      'int': int,
                      'float': float}

    def logic(self, context):
        sequence = self.sequence(context)
        default = self.default(context)
        if default is None:
            obj = {}
        else:
            obj = defaultdict(self._default_types[default])
        obj.update(self.get_let_map(context))
        if sequence:
            obj.update(sequence)
        dst = self.set_context(context,
                               self.dst(context),
                               obj)

        with context.scope(dst):
            yield DeferNodeContents(self)


class MakeToken(DataSetter):
    """Generates a token of random characters. This is useful for creating unique IDs for database objects."""

    class Help:
        synopsis = """make an opaque token"""
        example = """
        <maketoken dst="authorization_token" size="16" />
        """

    lowercase = Attribute("Use lower case characters", type="boolean", default=True, required=False)
    uppercase = Attribute("Use upper case characters", type="boolean", default=True, required=False)
    digits = Attribute("Use digits", type="boolean", default=True, required=False)
    punctuation = Attribute("Use punctuation", type="boolean", default=False, required=False)
    size = Attribute("Size", type="integer", default=20, required=False)

    def get_value(self, context):
        (size,
         lowercase,
         uppercase,
         digits,
         punctuation) = self.get_parameters(context,
                                            "size",
                                            "lowercase",
                                            "uppercase",
                                            "digits",
                                            "punctuation")
        choices = ""
        if lowercase:
            choices += string.lowercase
        if uppercase:
            choices += string.uppercase
        if digits:
            choices += string.digits
        if punctuation:
            choices += string.punctuation
        if not choices:
            self.throw("token", "No characters to choice from")
        token = ''.join(choice(choices) for _ in xrange(size))
        return token


class ODict(Dict):
    """Like <dict> but creates an [i]ordered[/i] dictionary (maintains the order data was inserted)/"""

    def logic(self, context):
        dst = self.set_context(context,
                               self.dst(context),
                               OrderedDict(self.get_let_map(context)))
        context.push_scope(dst)
        yield DeferNodeContents(self)


class Throw(Dict):
    """Throw a Moya exception. An exception consists of a name in dotted notation with an optional message.

    Custom exceptions should be named to start with the [i]long name[/i] of the library. Moya reserves exceptions with a single dot for internal use. These [i]may[/i] be thrown if appropriate.

    """

    class Help:
        synopsis = """throw an exception"""
        example = """
        <try>
            <throw exception="middle.earth.cant" msg="One can't simply do that!"/>
        </try>
        <except exception="middle.earth.*" dst="error">
            <echo>Error: ${error.msg}</echo>
        </except>

        """

    exception = Attribute("Type of exception", required=True)
    msg = Attribute("Message", default="exception thrown", required=False)

    def logic(self, context):
        exception, msg = self.get_parameters(context, 'exception', 'msg')
        info = context['_e'] = {}
        with context.scope('_e'):
            yield DeferNodeContents(self)
        raise MoyaException(exception, msg, info)


class Choose(DataSetter):
    """Pick a random element from a sequence."""

    class Help:
        synopsis = """choose a random value from a collection"""
        example = """
        <choose dst="hobbit" from="['bilbo','sam','frodo']"/>
        <echo>Hobbit: ${hobbit}</echo>
        """
    _from = Attribute("Container", required=True, type="expression")
    dst = Attribute("Destination", default=None, metavar="DESTINATION", type="reference", required=True)

    def logic(self, context):
        _from, dst = self.get_parameters(context, 'from', 'dst')
        self.set_context(context, dst, choice(_from))


class JSON(DataSetter):
    """Sets data from JSON (JavasSript Object Notation)."""

    src = Attribute("JSON source", type="expression", required=False)

    class Help:
        synopsis = """create data from JSON"""
        example = """
        <json dst="hobbits">
            ["Bilbo", "Sam", "Frodo"]
        </json>

        """

    def logic(self, context):
        self.set_context(context,
                         self.dst(context),
                         self.get_value(context))

    def get_value(self, context):
        if self.has_parameter('src'):
            text = self.src(context)
        else:
            text = self.text
        try:
            return json.loads(text)
        except Exception as e:
            self.throw('json.invalid', text_type(e))


# class UpdateURLquery(DataSetter):
#     """Update the query portion of a URL with data from the [i]let map[/i]."""

#     url = Attribute("URL", required=True)

#     def logic(self, context):
#         url = self.url(context)

#         parsed_url = urlparse(url)

#         url_query = parsed_url.query
#         query_components = parse_qs(url_query)
#         query_components.update(self.get_let_map(context))

#         url = urlunparse((parsed_url.scheme,
#                           parsed_url.netloc,
#                           parsed_url.path,
#                           parsed_url.params,
#                           urlencode(query_components, True),
#                           parsed_url.fragment))
#         self.set_context(context,
#                          self.dst(context),
#                          url)


class Slice(ContextElementBase):
    """Slice a sequence of values (extract a range of indexed values)."""

    class Help:
        synopsis = """get a 'slice' of a collection"""
        example = """
        <slice src="hobbits" start="1" stop="3" dst="slice_of_hobbits"/>

        """

    src = Attribute("Value to slice", required=True, type="reference", metavar="ELEMENTREF")
    dst = Attribute("Destination for slice", required=True, type="reference", metavar="ELEMENTREF")
    start = Attribute("Start point", type="expression", default=None)
    stop = Attribute("End point", type="expression", default=None)

    def logic(self, context):
        src, dst, start, stop = self.get_parameters(context, 'src', 'dst', 'start', 'stop')
        src_obj = context[src]

        if hasattr(src_obj, 'slice'):
            context[dst] = src_obj.slice(start or 0, stop)
        else:
            context[dst] = src_obj[start:stop]


class Page(ContextElementBase):
    """Chop a sequence in to a [i]page[/i].

    This breaks up a sequence in to a page, based on a page size and index. Useful for paginating search results.
    """

    class Help:
        synopsis = """get a page of results"""
        example = """
        <page src="results" page="2" pagesize="10" />
        """

    src = Attribute("Value to slice", required=True, type="reference", metavar="ELEMENTREF")
    dst = Attribute("Destination for slice", required=True, type="reference", metavar="ELEMENTREF")
    page = Attribute("Page", type="expression", required=False, default=".request.GET.page or 1", evaldefault=True)
    pagesize = Attribute("Page size", type="expression", required=False, default=10)

    def logic(self, context):
        src, dst, page, pagesize = self.get_parameters(context, 'src', 'dst', 'page', 'pagesize')
        try:
            page = int(page or 1)
        except ValueError:
            page = 1
        start = (page - 1) * pagesize
        stop = page * pagesize
        src_obj = context[src]
        if hasattr(src_obj, 'slice'):
            context[dst] = src_obj.slice(start, stop)
        else:
            context[dst] = src_obj[start:stop]


class Inc(ContextElementBase):
    """Increment (add 1 to) a value."""

    class Help:
        synopsis = """increment a value"""
        example = """
        <let hobbit_count="1" />
        <inc dst="hobbit_count" />
        <echo>${hobbit_count}</echo> <!-- 2 -->
        """

    dst = Attribute("Value to increment", required=True, type="reference", metavar="ELEMENTREF")

    def logic(self, context):
        try:
            context[self.dst(context)] += 1
        except:
            self.throw('inc.fail',
                       'unable to increment')


class Dec(ContextElementBase):
    """Decrement (subtract 1 from) a value."""

    class Help:
        synopsis = """decrement a value"""
        example = """
        <let hobbit_count="2" />
        <dec dst="hobbit_count" />
        <echo>${hobbit_count}</echo> <!-- 1 -->
        """
    dst = Attribute("Value to decrement", required=True, type="reference")

    def logic(self, context):
        try:
            context[self.dst(context)] -= 1
        except:
            self.throw('dec.fail',
                       'unable to decrement')


class Copy(ContextElementBase):
    """Copy a variable from one location to another"""

    class Help:
        synopsis = """copy a value"""
        example = """
        <str dst="foo">Kirk</str>
        <copy src="foo" dst="bar" />
        <echo>${bar}</echo>
        """

    src = Attribute("Source", required=True, type="reference")
    dst = Attribute("Destination", required=True, type="reference")

    def logic(self, context):
        context.copy(*self.get_parameters(context, 'src', 'dst'))


class Link(ContextElementBase):
    """Create a [i]link[/i] in the context. Similar to a symlink in the filesystem, a link creates a variable that references another value."""

    class Help:
        synopsis = """link variables"""
        example = """
        <str dst="foo">Hello</str>
        <link src="foo" dst="bar" />
        <str dst="bar">World</str>
        <echo>${foo}</bar> <!-- World -->
        <echo>${bar}</bar> <!-- World -->

        """

    src = Attribute("Source", required=True, type="reference")
    dst = Attribute("Destination", required=True, type="reference")

    def logic(self, context):
        context.link(*self.get_parameters(context, 'dst', 'src'))


class Append(ContextElementBase):
    """Append a value to a list or other collection."""

    class Help:
        synopsis = """append a value to a list"""
        example = """
        <list dst="crew">
            <str>John</str>
        </list>
        <append src="crew" value="'Scorpius'"/>
        """

    class Meta:
        one_of = [('value', 'values')]

    value = Attribute("Value to append", type="expression", missing=False, required=False)
    values = Attribute("A sequence of values to append", type="expression", missing=False, required=False)
    src = Attribute("Collection to append to", type="expression", required=True, missing=False)

    def logic(self, context):
        value, values, src = self.get_parameters(context, 'value', 'values', 'src')
        if self.has_parameter('value'):
            values = [value]
        for v in values:
            try:
                src.append(v)
            except:
                if not hasattr(src, 'append'):
                    self.throw("bad-value.not-supported",
                               "src does not support append operation",
                               diagnosis="Not all objects may be appended to, try using a list or list-like object.")
                else:
                    self.throw("append.fail",
                               "unable to append",
                               diagnosis="Check the value you are appending is the type expected.")


class Remove(ContextElementBase):
    """Remove a value from a collection"""

    class Help:
        synopsis = """remove a value from a collection"""
        example = """
        <list dst="crew">
            <str>John</str>
            <str>Scorpius/str>
        </list>
        <remove src="crew" value="'Scorpius'"/>
        """

    class Meta:
        one_of = [('value', 'values')]

    value = Attribute("Value to remove", type="expression", missing=False, required=False)
    values = Attribute("A sequence of values to remove", type="expression", missing=False, required=False)
    src = Attribute("Collection to remove from", type="expression", required=True, missing=False)

    def logic(self, context):
        value, values, src = self.get_parameters(context, 'value', 'values', 'src')
        if value is not None:
            values = [value]
        for v in values:
            try:
                src.remove(v)
            except ValueError:
                # removing non-existent value is a nop
                continue
            except:
                if not hasattr(src, 'remove'):
                    self.throw("bad-value.not-supported",
                               "src does not support remove operation")
                else:
                    self.throw("remove.fail",
                               "unable to remove")


class Extend(ContextElementBase):
    """Extend a sequence with values from another sequence."""

    class Help:
        synopsis = """add values from one sequence to another"""
        example = """
        <list dst="crew">
            <str>John</str>
        </list>
        <extend dst="crew" src="['Scorpius', 'Aeryn']"/>
        """

    src = Attribute("Collection to append to", type="expression")
    values = Attribute("value(s) to extend with", type="reference")

    def logic(self, context):
        src, values = self.get_parameters(context, 'src', 'values')
        try:
            src.extend(values)
        except:
            if not hasattr(src, 'extend'):
                self.throw('bad-value.not-supported',
                           'src does not support extend')
            else:
                self.throw('extend.fail',
                           'unable to extend')


class Update(ContextElementBase):
    """Update a dictionary (or other mapping collection) with keys and values from another dictionary."""

    class Help:
        synopsis = """add new values to a dictionary"""
        example = """
        <dict dst="species"/>
        <update dst="species" src="john='human', scorpius='unknown'"/>
        """

    src = Attribute("Collection to update", type="expression")
    values = Attribute("Values to update with", type="expression")

    def logic(self, context):
        src, values = self.get_parameters(context, 'src', 'values')

        try:
            src.update(values)
        except:
            if not hasattr(src, 'update'):
                self.throw('bad-value.not-supported',
                           "src does not support update")
            else:
                self.throw('update.fail',
                           'failed to update')


class Pop(DataSetter):
    """Remove and return a value from a dictionary."""

    class Help:
        synopsis = """remove a value from a dictionary"""
        example = """
        <dict dst="species">
            <let john="human"/>
            <let scorpius="unknown"/>
        </dict>
        <pop src="species" key="scorpius" dst="scorpius_species"/>
        <echo>${scorpius_species}</echo> <!-- "unknown" -->
        """

    src = Attribute("Source", type="expression")
    dst = Attribute("Destination", type="reference")
    key = Attribute("Key", required=False)

    def logic(self, context):
        src, dst, key = self.get_parameters(context, 'src', 'dst', 'key')
        try:
            if self.has_parameter('key'):
                value = src.pop(key)
            else:
                value = src.pop()
        except IndexError:
            self.throw('pop.empty',
                       "can't pop from an empty list")
        except:
            self.throw('pop.fail',
                       'unable to pop from src')

        self.set_context(context, dst, value)


class Echo(ContextElementBase):
    """Write text to the console. This is typically used in commands or for debugging. Bear in mind that in a production environment there is no console."""

    class Help:
        synopsis = """write text to the console"""
        example = """<echo>Hello, World</echo>"""

    obj = Attribute("If given, then the value of this expression will be written to the console",
                    type="expression", required=False, default=None)
    table = Attribute("A table to display. If given this value should be a list of lists.", type="expression", required=False, default=None)
    header = Attribute("A list of headers for the table (if given)", type="commalist", required=False, default=None)
    indent = Attribute("Number of indents", type="expression", required=False, default=0)
    style = Attribute("Console style", required=False, default="bold cyan")

    def logic(self, context):
        obj = self.obj(context)
        tabs = self.indent(context)
        table = self.table(context)
        header = self.header(context)
        style = console_style(self.style(context))
        console = context.root.get('console', None)
        if console is None:
            from .. import pilot
            console = pilot.console
        if self.archive.debug_echo:
            line_msg = '[file "{}", line {}]\n'.format(self._location, self.source_line)
            console(line_msg, fg="yellow")
        if self.has_parameter('table'):
            console.table(table, header_row=header)
        else:
            if obj is not None:
                console.obj(context, obj, **style)
            else:
                console.text(context.sub(("    " * tabs) + self.text), **style)


class Exit(ContextElementBase):
    """Exit the current command. When Moya sees this tag it stops running the current command and returns an exit code. It will also write any enclosed text to the console."""

    class Help:
        synopsis = """exit the current command"""
        example = """<exit code="-5" if="not continue">User cancelled</exit>"""

    code = Attribute("Return code", type="expression", default=0, map_to="exit_code")

    def logic(self, context):
        exit_code = self.exit_code(context)
        if self.text.strip():
            context.root['console'].text(context.sub(self.text.strip()), fg="cyan", bold=True)
        raise EndLogic(exit_code)


class Debug(ContextElementBase):
    """Executes a block only if Moya is in debug mode. You can enable debug mode the setting 'debug' under the project section. This tag is equivalent to the following:

[code xml]
<if test=".debug">
    <echo>Moya is in debug mode. Relax.</echo>
</if>
[/code]

[alert]Be careful to not do anything inside this tag that your project may depend on. It will not execute in a production environment![/alert]

    """

    class Help:
        synopsis = """execute debug code"""
        example = """
        <debug>
            <echo>Moya is in debug mode. Relax.</echo>
        </debug>
        """

    def logic(self, context):
        if context.get('.debug', False):
            yield DeferNodeContents(self)


class For(ContextElementBase):
    """Execute a block of code for each value in a sequence. This tag [i]iterates[/i] over a sequence, and executes the enclosed block for each value."""

    class Help:
        synopsis = "iterate over a sequence"
        example = """
        <for src="['John', 'Scorpius', 'Rygel']" dst="crew_member" filter="crew_member != 'Rygel'">
            <echo>${crew_member} is on board!</echo>
        </for>

        """

    class Meta:
        is_loop = True

    src = Attribute("Source sequence", required=True, type="expression")
    dst = Attribute("Destination value", required=True, type="commalist")
    filter = Attribute("If given, then only those values which match this condition will cause the enclosed block to execute.",
                       required=False, type="expression", default=True)

    def logic(self, context, _defer=DeferNodeContents):
        objects, dst = self.get_parameters(context, 'src', 'dst')

        filter = self.filter
        try:
            iter_objects = iter(objects)
        except TypeError:
            self.throw("bad-value.not-iterable", "source is not iterable")
        else:
            if len(dst) == 1:
                dst = dst[0]
                for obj in iter_objects:
                    context[dst] = obj
                    if filter(context):
                        yield _defer(self)
            else:
                for obj in iter_objects:
                    try:
                        for set_dst, set_value in zip(dst, obj):
                            context[set_dst] = set_value
                    except TypeError:
                        self.throw("bad-value.not-iterable", "Object in sequence does not support iteration")
                    if filter(context):
                        yield _defer(self)


class Switch(ContextElementBase):
    """
    Jump to a location with in a code block, based in an input value.

    This tag takes a value ([c]on[/c]) and jumps to a [tag]case[/tag] in the enclosed block that has a matching value.
    If there is no matching [tag]case[/tag], then execution will move to the enclosed [tag]default-case[/tag]. If no default case exists, the entire block is skipped.

    This tag is useful in situations where a chain of [tag]if[/tag] and [tag]elif[/tag] tags would otherwise be used. Here's an example:

    [code xml]
    <switch on="species">
        <case>human</case>
        <echo>Humans come from Earth</echo>

        <case>hynerian</case>
        <echo>Hynerians come from Hyneria</echo>

        <default-case/>
        <echo>I have no idea where ${species}s come from</echo>
    </switch>
    [/code]

    The above code is equivalent to the following:

    [code xml]
    <if test="species=='human'">
        <echo>Humans come from Earth</echo>
    </if>
    <elif test="species=='hynerian'">
        <echo>Hynerians come from Hyneria</echo>
    </elif>
    <else>
        <echo>I have no idea where ${species}s come from</echo>
    </else>
    [/code]

    The [tag]switch[/tag] version is generally clearer, especially with a large number of conditions.

    If the [c]on[/c] attribute is [i]not[/i] given, then the case's [c]if[/c] attribute is checked. Here's the previous example, implemented without the [c]on[/c] attribute:

    [code xml]
    <switch>
        <case if="species=="'human'"/>
        <echo>Humans come from Earth</echo>

        <case if="species=="'hynerian'"/>
        <echo>Hynerians come from Hyneria</echo>

        <default-case/>
        <echo>I have no idea where ${species}s come from</echo>
    </switch>
    [/code]

    """

    class Help:
        synopsis = "jump to matching case"

    class Meta:
        is_loop = True

    on = Attribute("Value to test", type="expression", required=False)

    def logic(self, context):
        case_tags = [(namespaces.default, 'case'), (namespaces.default, 'default-case')]

        if self.has_parameter('on'):
            switch_on = self.on(context)
            iter_children = iter(self)
            while 1:
                child = next(iter_children, None)
                if child is None:
                    return
                if child._element_type in case_tags and child.test(context, switch_on):
                    for child in iter_children:
                        yield child
                    break
        else:
            iter_children = iter(self)
            while 1:
                child = next(iter_children, None)
                if child is None:
                    return
                if child._element_type in case_tags and child._if(context):
                    for child in iter_children:
                        yield child
                    break


class Case(ContextElementBase):
    """
    Defines a case in a [tag]switch[/tag] tag.

    If a [c]value[/c] attribute is given, it will be used as the comparison values, otherwise the tag text will be used.

    """

    class Help:
        synopsis = "define a case in a switch"

    value = Attribute("Value to compare in a switch", type="expression", required=False)

    def check(self, context):
        return True

    def test(self, context, switch_value):
        if self.has_parameter('value'):
            value = self.value(context)
        else:
            value = context.sub(self.text.strip())
        return value == switch_value

    def logic(self, context):
        raise BreakLoop()


class DefaultCase(ContextElementBase):
    """
    Defines the [i]default[/i] case in a [tag]switch[/tag] tag.

    """

    class Help:
        synopsis = "default case in a switch"

    def check(self, context):
        return True

    def test(self, context, switch_value):
        return True

    def logic(self, cotnext):
        raise BreakLoop()


class ProgressElement(ContextElementBase):
    """Like a for loop but renders an ascii progress bar in the console, which will look something like the following:

    [code]
    [###        ] 30% working...
    [/code]

    This is useful for commands that may take some time to execute.

    """

    class Help:
        synopsis = """render an ascii progress bar"""

    src = Attribute("Source", required=True, type="expression")
    dst = Attribute("Destination", required=True, type="commalist")
    msg = Attribute("Message on progress bar", required=False, default="working...")
    steps = Attribute("Number of steps in the sequence", required=False, type="integer")

    class Meta:
        tag_name = "progress"
        is_loop = True

    def logic(self, context):
        objects, dst, msg, steps = self.get_parameters(context, 'src', 'dst', 'msg', 'steps')
        console = context.root['console']

        if steps is None:
            try:
                steps = len(objects)
            except:
                self.throw('moya.progress.no-length',
                           "Unable to get length of {!r}".format(objects))

        progress = Progress(console, msg, num_steps=steps)
        progress.render()

        try:
            iter_objects = iter(objects)
        except TypeError:
            self.throw("bad-value.not-iterable", "Source is not iterable")
        else:
            try:
                console.show_cursor(False)
                if len(dst) == 1:
                    dst = dst[0]
                    for obj in iter_objects:
                        context[dst] = obj
                        progress.step(msg=self.msg(context))
                        yield DeferNodeContents(self)


                else:
                    for obj in iter_objects:
                        try:
                            for set_dst, set_value in zip(dst, obj):
                                context[set_dst] = set_value
                        except TypeError:
                            self.throw("bad-value.not-iterable",
                                       "Object in sequence does not support iteration")
                        yield DeferNodeContents(self)
                        progress.step()
            finally:
                console.show_cursor(True)
                progress.done()


class Sleep(ContextElementBase):
    """Do nothing for a period of time"""

    class Help:
        synopsis = """wait for while"""
        example = """<sleep for="10s"/> <!-- do nothing for 10 seconds -->"""

    _for = Attribute("Time", required=True, type="timespan")

    def logic(self, context):
        t = self.get_parameter(context, 'for')
        sleep(float(t))


class Map(DataSetter):
    """Create a list by mapping an expression on to a sequence"""

    class Help:
        synopsis = "map an expression on to sequence"
        example = """
        <dict dst="crew">
            <dict let:name="john" let:species="human" />
            <dict let:name="rygel" let:species="hynerian" />
            <dict let:name="aeryn" let:species="peacekeeper" />
        </dict>
        <map src="crew" dst="manifest"
            expression="sub:'${title:name$} is ${species}'" />
        <!-- ['John is human', 'Rygen is hynerian', 'Aeryn is peacekeeper'] -->
        """

    src = Attribute("Source sequence", required=True, type="expression")
    dst = Attribute("Destination", required=False, type="reference")
    value = Attribute("Expression", required=True, type="function")
    _filter = Attribute("Skip item if this expression is false", required=False, type="function")

    def get_value(self, context):
        objects, dst, func, _filter = self.get_parameters(context, 'src', 'dst', 'value', 'filter')
        func = func.get_scope_callable(context)
        if _filter is None:
            result = [func(obj) for obj in objects]
        else:
            _filter = _filter.get_scope_callable(context)
            result = [func(obj) for obj in objects if _filter(obj)]

        return result


class MapDict(DataSetter):
    """Create a list of dictionaries from a sequence."""

    class Help:
        synopsis = "generate a list of dictionaries"
        example = """
        <dict dst="crew">
            <dict let:name="john" let:species="human" />
            <dict let:name="rygel" let:species="hynerian" />
            <dict let:name="aeryn" let:species="peacekeeper" />
        </dict>
        <map-dict src="crew" dst="crew" let:name="title:name" let:human="species == 'human'"/>
        <!-- [{'name':John, 'human':yes}, {'name':Rygel, 'human':no}, {'name':'Aeryn',  'human':no}] -->
        """

    src = Attribute("Source sequence", required=True, type="expression")
    dst = Attribute("Destination", required=False, type="reference")
    _filter = Attribute("Skip item if this expression is false", required=False, type="function")

    def logic(self, context):

        objects, dst, _filter = self.get_parameters(context, 'src', 'dst', 'filter')

        if _filter is None:
            predicate = lambda obj: True
        else:
            predicate = _filter.get_scope_callable(context)

        map_result = []
        for obj in objects:
            with context.data_scope(obj):
                if not predicate(obj):
                    continue
                value = self.get_let_map(context)
                with context.data_scope(value):
                    yield DeferNodeContents(self)
                map_result.append(value)
        self.set_context(context,
                         dst,
                         map_result)


class Max(DataSetter):
    """Get the maximum value in a sequence."""

    class Help:
        synopsis = "get the maximum value in a sequence"

    src = Attribute("Source sequence", required=True, type="expression")
    key = Attribute("Key", type="function", required=False, default=None, missing=False)

    def logic(self, context):
        objects, dst, key = self.get_parameters(context, 'src', 'dst', 'key')
        if not objects:
            self.throw('max.empty', 'src is empty', diagnosis="Moya can't calculate the maximum of an empty sequence.")
        if key is None:
            result = max(objects)
        else:
            key_callable = key.get_scope_callable(context)
            result = max(key_callable(obj) for obj in objects)
        self.set_context(context, dst, result)


class Min(DataSetter):
    """Get the minimum value in a sequence."""

    class Help:
        synopsis = "get the minimum value in a sequence"

    src = Attribute("Source sequence", required=True, type="expression")
    key = Attribute("Key", type="function", required=False, default=None, missing=False)

    def logic(self, context):
        objects, dst, key = self.get_parameters(context, 'src', 'dst', 'key')
        if not objects:
            self.throw('min.empty', 'src is empty', diagnosis="Moya can't calculate the minimum of an empty sequence.")
        if key is None:
            result = min(objects)
        else:
            key_callable = key.get_scope_callable(context)
            result = min(key_callable(obj) for obj in objects)
        self.set_context(context, dst, result)


class FilterSeq(DataSetter):
    """Filter members of a collection which don't pass a condition."""

    class Help:
        synopsis = "filter values from a sequence"
        example = """
        <dict dst="crew">
            <dict let:name="john" let:species="human" />
            <dict let:name="rygel" let:species="hynerian" />
            <dict let:name="aeryn" let:species="peacekeeper" />
        </dict>
        <filter-seq src="items:crew dst="crew" test="species !- 'hyneria'"/>
        <!-- [{'name': 'john', 'species':'human'}, {'name':'aeryn', 'species':'peacemaker'}] -->

        """

    src = Attribute("Source", required=True, type="expression")
    dst = Attribute("Destination", required=False, type="reference", default=None)
    test = Attribute("Condition", required=True, type="function")

    def get_value(self, context):
        objects, dst, func = self.get_parameters(context, 'src', 'dst', 'test')
        predicate = func.get_scope_callable(context)
        return [item for item in objects if predicate(item)]


class Sort(DataSetter):
    """Sort a sequence"""

    class Help:
        synopsis = "sort a sequence"
        example = """
        <dict dst="crew">
            <dict let:name="John" let:species="human" />
            <dict let:name="Rygel" let:species="hynerian" />
            <dict let:name="Aeryn" let:species="peacekeeper" />
        </dict>
        <sort src="crew" dst="crew" key="lower:name" />
        """

    src = Attribute("Source sequence", required=True, type="expression")
    dst = Attribute("Destination", required=False, type="reference", default=None)
    key = Attribute("Key expression", required=True, type="function")
    reverse = Attribute("Reverse order?", type="boolean", required=False, default=False)

    def get_value(self, context):
        objects, dst, func, reverse = self.get_parameters(context, 'src', 'dst', 'key', 'reverse')
        get_key = func.get_scope_callable(context)
        return sorted(objects,
                      reverse=reverse,
                      key=get_key)


class Repeat(ContextElementBase):
    """Repeat a block a set number of times, or indefinitely. You can leave a repeat loop prematurely with the [tag]break[/tag] tag.

    [alert Note]Be careful not to create infinite loops with this tag[/alert]

    """

    class Help:
        synopsis = """repeat a block"""

        example = """
        <repeat times="5">
            <echo>Candyman</echo>
        </repeat>
        """

    times = Attribute("Number of times to repeat (default is infinite)", type="expression", default=None)

    class Meta:
        is_loop = True

    def logic(self, context):
        times = self.times(context)
        if times is not None:
            try:
                times = int(times)
            except:
                self.throw("bad-value.not-a-number", "'times' must be a number if given")
        if times is None:
            while 1:
                yield DeferNodeContents(self)
        else:
            count = max(0, times)
            while count:
                yield DeferNodeContents(self)
                count -= 1


class While(ContextElementBase):
    """Repeat a block of code while a condition is true, or a [tag]break[/tag] tag is reached.

    [alert Note]Be careful not to create an [i]infinite[/i] loop with this tag.[/alert]

    """

    class Help:
        synopsis = "repeat a block while a condition is true"
        example = """

        <let i="5"/>
        <while test="i gt 0">
            <echo>${i}</echo>
            <let i="i-1"/>
        </while>


        """

    class Meta:
        is_loop = True

    test = Attribute("Condition", required=True, type="expression")

    def logic(self, context, _defer=DeferNodeContents):
        test = self.test
        while test(context):
            yield _defer(self)


class Break(ContextElementBase):
    """This tag will [i]break[/i] a loop such as [tag]while[/tag], [tag]for[/tag]. When Moya encounters this tag, it jumps to the statement after the loop."""

    class Help:
        synopsis = """end a loop prematurely"""
        example = """

        <let crew="['John', 'Rygel', 'Scorpius', 'Aeryn']"/>
        <for src="crew" dst="character">
            <if test="character == 'Scorpius'">
                <echo>Taking off before Scorpius gets on board!</eco>
                <break>
            </if>
            <echo>${character} is on board</echo>
        </for>

        """

    def logic(self, context):
        raise BreakLoop()


class Continue(ContextElementBase):
    """When Moya encounters this tag in a loop, such as [tag]while[/tag] or [tag]for[/tag], it ignores the remaining code in the block and [i]continues[/i] to the next item in the loop."""

    class Help:
        synopsis = """skip remaining code in a loop"""
        example = """

        <let crew="['John', 'Rygel', 'Scorpius', 'Aeryn']"/>
        <for src="crew" dst="character">
            <if test="character == 'Scorpius'">
                <echo>Scorpius is not allowed on board!</eco>
                <continue/>
            </if>
            <echo>${character} is on board</echo>
        </for>


        """

    def logic(self, context):
        raise ContinueLoop()


class Macro(ContextElementBase):
    """Defines a [link moyacode#macros]macro[/link]."""

    class Help:
        synopsis = """define a re-usable block of code"""
        example = """
        <macro docname="greet">
            <echo>Hello, ${name}!<echo>
        </macro>
        """

    def finalize(self, context):
        for signature in self.children('signature'):
            self.validator = signature.validator
            self.validate_call = self.validator.validate


class Preflight(Macro):
    """A pre-flight check is used to detect potential problems, such as missing settings."""

    class Help:
        synopsis = """check initial settings"""


class ReturnDict(ContextElementBase):
    """
    Shortcut to return a dictionary. For example, the following macro returns a dictionary:

    [code xml]
    <macro docname="get_character">
        <return>
            <dict>
                <str dst="rygel">Hynerian</str>
                <str dst="john">Human</str>
            </dict>
        </return>
    </macro>
    [/code]

    We could shorten the above using the [tag]return-dict[/tag] as follows:

    [code xml]
    <macro docname="get_character">
        <return-dict>
            <str dst="rygel">Hynerian</str>
            <str dst="john">Human</str>
        </return-dict>
    </macro>
    [/code]

    """

    class Help:
        synopsis = "shortcut to return a dictionary"

    def logic(self, context):
        data = self.get_let_map(context).copy()
        context['_return'] = data
        with context.scope('_return'):
            yield DeferNodeContents(self)
        raise Unwind()


class ReturnScope(ContextElementBase):
    """
    Return values from the current scope.

    This tag is useful in macros, widgets, and other callable tags where you want to return a number of values in a dictionary.

    For example, the following will return a dictionary containing two keys:

    [code xml]
    <return-scope values="foo, bar"/>
    [/code]

    This is equivalent to the following:

    [code xml]
    <return value="{'foo': foo, 'bar': bar}"/>
    [/code]

    If you *don't* specify the [c]values[/c] attribute, Moya will read the values from the enclosed text (one key per line).
    The following code is equivalent to the above:

    [code xml]
    <return-scope>
        foo
        bar
    </return-scope>
    [/code]

    """

    values = Attribute("Values to return", type="commalist", required=False, default=None)
    default = Attribute("Default for missing values", type="expression", required=False, default=None)

    class Help:
        synopsis = "return values from the current scope"

    def logic(self, context):
        names, default = self.get_parameters(context, 'values', 'default')
        if names is None:
            names = [l.strip() for l in context.sub(self.text).splitlines() if not l.isspace()]
        get = context.get
        context['_return'] = {name: get(name, default) for name in names}
        raise Unwind()


class ReturnStr(ContextElementBase):
    """A shortcut to returning a string."""

    class Help:
        synopsis = "shortcut to return a string"
        example = """
        <macro docname="get_text">
            <return-str>Hello, World</return-str>
        </macro>
        """

    def logic(self, context):
        context['_return'] = context.sub(self.text)
        raise Unwind()


class ReturnFalse(ContextElementBase):
    """
    A shortcut to return False.

    The following two lines are equivalent:

    [code xml]
    <return-false/>
    <return value="no" />
    [/code]

    """

    class Help:
        synopsis = "shortcut to return false"

    def logic(self, context):
        context['_return'] = False
        raise Unwind


class ReturnTrue(ContextElementBase):
    """
    A shortcut to return True.

    The following two lines are equivalent:

    [code xml]
    <return-true/>
    <return value="yes" />
    [/code]

    """

    class Help:
        synopsis = "shortcut to return true"

    def logic(self, context):
        context['_return'] = True
        raise Unwind


class Return(ContextElementBase):
    """
    Used in a callable block such as a [tag]macro[/tag] or [tag]view[/tag] to return data.

    If you enclose any [i]data setter[/i] tag ([tag]int[/tag], [tag]str[/tag], [tag]list[/tag] etc.) in the return block, the result will be returned. For example, the following macro will return a list of three items:

    [code xml]
    <macro libname="get_crew">
        <return>
            <list>
                <str>John</str>
                <str>Scorpius</str>
                <str>Rygel</str>
            </list>
        </return>
    </macro>
    [/code]

    Alternatively, you may use the [c]value[/c] attribute to return the result of an expression. The following code is equivalent to the above:

    [code xml]
    <macro libname="get_crew">
        <return value="['John', 'Scorpius', 'Rygel']" />
    </macro>
    [/code]

    """

    class Help:
        synopsis = """return data from a macro or other callable"""

    value = Attribute("Value to return", type="expression", default=None)

    def logic(self, context):
        if self.has_parameter('value'):
            context['_return'] = ReturnContainer(value=self.value(context))
        else:
            context['_return'] = ReturnContainer()
            with context.scope('_return'):
                yield DeferNodeContents(self)
        raise Unwind()


class CacheReturn(ContextElementBase):
    """
    Return a value from a cache, if it exists. Otherwise, execute the enclosed block.

    This tag can be used to [i]memoize[/i] a [tag]macro[/tag] or other callable. Essentially, this means that if you call the macro a second time with the same parameters, it returns the previously calculated result. For macros that are slow to execute, this can result in significant speedups.

    For example, the following code calculates the [url http://en.wikipedia.org/wiki/Factorial]factorial[/url] of a number:


    [code xml]
    <moya xmlns="http://moyaproject.com"
        xmlns:let="http://moyaproject.com/let">

        <macro docname="fact">
            <signature>
                <argument name="n"/>
            </signature>
            <cache-return key="n">
                <echo>calculating ${n}!</echo>
                <let f="1"/>
                <while test="n">
                    <let f="f*n" n="n-1"/>
                </while>
                <return value="f"/>
            </cache-return>
        </macro>

        <macro docname="main">
            <call macro="fact" let:n="7" dst="result"/>
            <echo>${result}</echo>
            <call macro="fact" let:n="7" dst="result"/>
            <echo>${result}</echo>
            <call macro="fact" let:n="7" dst="result"/>
            <echo>${result}</echo>
        </macro>

    </moya>
    [/code]

    If you run the above code, you will get the following output:

    [code]
    $ moya run cachereturn.xml
    calculating 7!
    5040
    5040
    5040
    [/code]

    The first time the [c]fact[/c] macro is called, Moya displays "calculating 7!" in the terminal. The second and third time, the text is [i]not[/i] displayed because the result is retrieved from the cache -- without the need to execute the code within [tag]cache-return[/tag].


    """

    class Help:
        synopsis = "cache returned values"

    cache = Attribute("Cache name", required=False, default="runtime")
    _for = Attribute("Time to cache for", required=False, default=0, type="timespan")
    key = Attribute("Cache key", required=False)
    keydata = Attribute("Cache data", type="expression", required=False)
    local = Attribute('Should the value be cached for this tag only?', type="boolean", default=True)

    class Meta:
        is_call = True
        one_of = [('key', 'keydata')]

    def logic(self, context):
        (cache_name,
         cache_time,
         cache_key,
         cache_key_data,
         cache_local) = self.get_parameters(context,
                                            'cache',
                                            'for',
                                            'key',
                                            'keydata',
                                            'local')
        if cache_key is None:
            cache_key = make_cache_key(cache_key_data)
        if cache_local:
            cache_key = "{}--{}".format(self.libid, cache_key)
        cache = self.archive.get_cache(cache_name)
        cache_result = cache.get(cache_key, Ellipsis)
        if cache_result is Ellipsis:
            call = context.get('.call', {})
            yield DeferNodeContents(self)
            if '_return' in call:
                value = _return = call['_return']
                if hasattr(_return, 'get_return_value'):
                    value = _return.get_return_value()
            else:
                value = None
            cache.set(cache_key, value, time=int(cache_time))
            context['_return'] = ReturnContainer(value=value)
            raise Unwind()

        else:
            context['_return'] = ReturnContainer(value=cache_result)
            raise Unwind()


class Done(ContextElementBase):
    """Exists the current callable immediately with no return value. Note, this is not equivalent to [tag]return[/tag] which returns [c]None[/c]."""

    class Help:
        synopsis = """return from a macro or callable without any data"""

    def logic(self, context):
        raise Unwind()


class Input(DataSetter):
    """
    Get data from the console.

    [alert WARNING]This tag should only be used in an interactive context, such as a [tag]command[/tag] tag.[/alert]

    """

    class Help:
        synopsis = "ask the user for information in a command"
        example = """
        <input dst="name">What is your name?</input>

        """

    default = ''
    _default = Attribute("Value to use if response is empty", default="", map_to="default")
    password = Attribute("Use password input?", default=False, type="boolean")

    def logic(self, context):
        default = self.default(context)
        text = context.sub(self.text) or ''
        if default:
            text += " ({})".format(default)
        if text:
            text += ' '
        if self.password(context):
            console = context.root['console']
            console(text).flush()
            response = getpass.getpass('')
        else:
            response = raw_input(text)
        if not response:
            response = default
        self.set_context(context,
                         self.dst(context),
                         response)


class Ask(DataSetter):
    """
    Ask the user a yes / no question.

    [alert WARNING]This tag should only be used in an interactive context, such as a [tag]command[/tag] tag.[/alert]

    """

    class Help:
        synopsis = "ask a yes / no question"
        example = """
        <ask dst="take_off">Would you like to take off?</ask>
        <echo if="take_off">Taking off!</echo>
        """

    default = False

    def logic(self, context):
        response = raw_input("%s (Y/N) " % (self.text.strip() or ''))
        response_bool = response.strip().lower() in ('yes', 'y')
        self.set_context(context,
                         self.dst(context),
                         response_bool)


class _LazyCallable(object):
    def __init__(self, context, _callable, args, kwargs):
        self.context = weakref.ref(context)
        self.callable = _callable
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        context = self.context()
        try:
            with context.frame():
                if getattr(self.callable, '_require_context', False):
                    result = self.callable(context, *self.args, **self.kwargs)
                else:
                    result = self.callable(*self.args, **self.kwargs)
        except Exception as e:
            log.exception("lazy error %s", e)
            raise
        else:
            if hasattr(result, 'get_return_value'):
                result = result.get_return_value()
            return result


class Call(ContextElementBase):
    """Call a [link moyacode#macros]macro[/link]."""

    class Help:
        synopsis = "invoke a macro"
        example = '<call macro="moya.auth#login" />'

    class Meta:
        is_call = True

    macro = Attribute("Macro", required=True)
    dst = Attribute("Destination for return value", type="reference")
    lazy = Attribute("If True the result will be evaluated lazily (i.e. when next referenced)", type="boolean")
    _from = Attribute("Application", default=None, type="application")

    def logic(self, context):
        macro, dst, lazy, app = self.get_parameters(context, 'macro', 'dst', 'lazy', 'from')
        items_map = self.get_let_map(context)
        app = app or self.get_app(context, check=False)

        if self.has_children:
            call = self.push_funccall(context)
            call.update(items_map)
            try:
                yield DeferNodeContents(self)
            finally:
                self.pop_funccall(context)
            args, kwargs = call.get_call_params()
        else:
            call = {}
            args = []
            kwargs = items_map

        lazy_callable = None
        if lazy:
            if not dst:
                self.throw('call.missing-dst', "A value for 'dst' is required for lazy calls")
            macro_app, macro_element = self.get_element(macro, app)

            if hasattr(macro_element, 'validate_call'):
                macro_element.validate_call(context, macro_element, call)

            element_callable = self.archive.get_callable_from_element(macro_element,
                                                                      app=macro_app or app)
            lazy_callable = _LazyCallable(context, element_callable, args, kwargs)
            context.set_lazy(dst, lazy_callable)

        else:
            macro_app, macro_element = self.get_element(macro, app)
            if macro_element._meta.app_first_arg:
                call = {'args': [macro_app] + args}
            else:
                call = {}
                if args:
                    call['args'] = args
            call.update(kwargs)

            if hasattr(macro_element, 'validate_call'):
                macro_element.validate_call(context, macro_element, call)

            self.push_call(context, call, app=macro_app)
            try:
                yield DeferNodeContents(macro_element)
            finally:
                call = self.pop_call(context)
            if '_return' in call:
                value = _return = call['_return']
                if hasattr(_return, 'get_return_value'):
                    value = _return.get_return_value()
            else:
                value = None
            if dst is None:
                getattr(context.obj, 'append', lambda a: None)(value)
            else:
                context[dst] = value


class CallElement(ContextElementBase):
    """
    Call a element object.

    This tag calls an element retrieved with tags such as [tag]get-element[/tag]. For a general purpose call tag, see [tag]call[/tag].

    """

    class Help:
        synopsis = "invoke a macro element"
        example = """
        <get-element name="sushifinder#macro.check_stock" dst="element" />
        <call-element element="element" dst="result"/>

        """

    class Meta:
        is_call = True

    element = Attribute("An element object", type="expression", required=True)
    dst = Attribute("Destination for return value", type="reference")

    def logic(self, context):
        macro, dst = self.get_parameters(context, 'element', 'dst')

        try:
            macro_element = macro.__moyaelement__()
            macro_app = macro.app
        except:
            self.throw('call-element.not_element',
                       'must be called with an element object')

        if self.has_children:
            call = self.push_funccall(context)
            try:
                yield DeferNodeContents(self)
            finally:
                self.pop_funccall(context)
            args, kwargs = call.get_call_params()
            call = {'args': args}
            call.update(kwargs)
            call.update(self.get_let_map(context))
        else:
            call = self.get_let_map(context)

        if hasattr(macro_element, 'validate_call'):
            macro_element.validate_call(context, macro_element, call)

        self.push_call(context, call, app=macro_app)
        try:
            yield DeferNodeContents(macro_element)
        finally:
            call = self.pop_call(context)
        if '_return' in call:
            value = _return = call['_return']
            if hasattr(_return, 'get_return_value'):
                value = _return.get_return_value()
        else:
            value = None
        if dst is None:
            getattr(context.obj, 'append', lambda a: None)(value)
        else:
            context[dst] = value


class Defer(ContextElementBase):
    """Defer to a given element, such as a [tag]macro[/tag]. [i]Deferring[/i] to a macro is similar to calling it, but no new [i]scope[/i] is created. This means that the macro has access to the same variables where defer was called."""

    class Help:
        synopsis = """jump to another element"""
        example = """
        <macro docname="board">
            <echo>${character} is on board</echo>
        </macro>

        <macro docname="main">
            <let crew="['John', 'Rygel', 'Scorpius']"/>
            <for src="crew" dst="character">
                <defer to="board"/>
            </for>
        </macro>

        """

    element = Attribute("Element", type="expression", required=False, default=None)
    _to = Attribute("Element reference", required=False, default=None, map_to="element_ref")
    _from = Attribute("Application", default=None, type="application")

    def logic(self, context):
        element, element_ref, app = self.get_parameters(context, 'element', 'element_ref', 'from')

        app = app or self.get_app(context)
        if element is not None:
            if not hasattr(element, '__moyaelement__'):
                self.throw("bad-value.not-an-element", "Can't defer to '{!r}' because it's not an element".format(element))
            element = element.__moyaelement__()
        elif element_ref is not None:
            app, element = self.get_element(element_ref, app)
        else:
            self.throw("bad-value.missing-element", "No value given for 'element' or 'element_ref'")
        if element._element_class != "logic":
            self.throw("defer.not-logic", "This element can not be deferred to")

        if element.has_children:
            with self.defer(context, app):
                yield DeferNodeContents(element)


class Scope(LogicElement):
    """Creates a new temporary [i]scope[/i] from an object."""

    class Help:
        synopsis = "create a temporary scope"
        example = """
        <dict dst="species">
            <let john="human"/>
            <let rygel="hynerian"/>
        </dict>
        <scope object="species">
            John is a ${john}, Rygel is a ${rygel}
        </scope>
        """

    object_ = Attribute("Object", type="expression", name="object", required=True)

    def logic(self, context):
        obj = self.object(context)
        with context.data_scope(obj):
            yield DeferNodeContents(self)



from __future__ import unicode_literals
from __future__ import print_function

from ..elements import Attribute
from ..interface import AttributeExposer
from ..tags.context import DataSetter

import requests

import logging
log = logging.getLogger('moya.runtime')


class ResponseProxy(AttributeExposer):
    """Proxy for a request object"""

    __moya_exposed_attributes__ = ["url",
                                   "text",
                                   "status_code",
                                   "headers",
                                   "cookies",
                                   "history",
                                   "content",
                                   "json",
                                   "encoding"]

    def __init__(self, req, url, method):
        self._req = req
        self._url = url
        self._method = method

    def __repr__(self):
        return '<httpresponse {} "{}">'.format(self._method, self._url)

    @property
    def url(self):
        return self._req.url

    @property
    def text(self):
        return self._req.text

    @property
    def status_code(self):
        return self._req.status_code

    @property
    def headers(self):
        return dict(self._req.headers)

    @property
    def cookies(self):
        return dict(self._req.cookies)

    @property
    def history(self):
        self._req.history

    @property
    def content(self):
        return self._req.content

    @property
    def json(self):
        try:
            return self._req.json()
        except:
            return None

    @property
    def encoding(self):
        return self._req.encoding


class RequestTag(DataSetter):
    """
    Make HTTP requests.

    """

    class Meta:
        tag_name = "request"

    class Help:
        synopsis = "make an http request"

    url = Attribute("URL to request", required=True)
    method = Attribute("Method to use", choices=['get', 'post', 'delete', 'put', 'trace', 'head', 'options'], default=None)
    params = Attribute("Request parameters", type="dict", required=False, default=None)
    headers = Attribute("Additional headers", type="dict", required=False, default=None)
    data = Attribute("Data to be form encoded", type="dict", required=False, default=None)
    timeout = Attribute("Timeout in seconds", type="number", required=False, default=None)

    username = Attribute("Username for basic auth", required=False, default=None)
    password = Attribute("Password for basic auth", required=False, default=None)

    def _get_method(self):
        return 'get'

    def logic(self, context):
        params = self.get_parameters(context)
        method = params.method or self._get_method()
        request_maker = getattr(requests, method)
        if not self.has_parameter('params'):
            request_params = self.get_let_map(context)
        else:
            request_params = params.params
        if params.username is not None:
            auth = (params.username, params.password or '')
        else:
            auth = None
        try:
            log.debug('requesting %s %s', method, params.url)
            response = request_maker(params.url,
                                     auth=auth,
                                     timeout=params.timeout,
                                     params=request_params,
                                     headers=params.headers,
                                     data=params.data)
        except requests.exceptions.Timeout:
            self.throw('requests.timeout',
                       'the server did not response in time',
                       diagnosos="Try raising the timeout attribute")
        except requests.exceptions.HTTPError:
            self.throw('requests.http-error',
                       'the server response was not invalid')
        except requests.exceptions.TooManyRedirects:
            self.throw('requests.too-many-redirects',
                       'too may redirect responses')
        except requests.exceptions.RequestException as e:
            self.throw('requests.error',
                       'unable to make request ({})'.format(e))

        response_proxy = ResponseProxy(response, params.url, method)
        self.set_context(context, params.dst, response_proxy)


class RequestGetTag(RequestTag):
    """
    Make a GET request (see [tag]request[/tag]).

    """

    class Help:
        synopsis = "make a get requests"

    class Meta:
        tag_name = "request-get"

    def _get_method(self):
        return 'get'


class RequestPostTag(RequestTag):
    """
    Make a POST request (see [tag]request[/tag]).

    """

    class Help:
        synopsis = "make a post request"

    class Meta:
        tag_name = "request-post"

    def _get_method(self):
        return 'post'



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .render import HTML
from . import html
from .compat import with_metaclass, implements_to_string, text_type
import postmarkup
import CommonMark


def get_installed_markups():
    return list(MarkupBaseMeta.markup_registry.keys())


class MarkupBaseMeta(type):
    markup_registry = {}

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        if name != "MarkupBase":
            name = getattr(new_class, 'name', name.lower().strip('_'))
            cls.markup_registry[name] = new_class
        return new_class


@implements_to_string
class MarkupBaseType(object):
    __metaclass__ = MarkupBaseMeta

    def __init__(self, markup_type, markup_options):
        self.markup_type = markup_type
        self.markup_options = markup_options
        self.create(markup_options)

    def __str__(self):
        return "<markup {}>".format(self.name)

    def create(self, options):
        pass

    @classmethod
    def get_processor(cls, name, markup_options):
        markup_class = MarkupBaseMeta.markup_registry[name]
        return markup_class(name, markup_options)

    def process(self, text, target, options):
        target = text_type(target or 'text')
        process_method = "process_" + target
        if hasattr(self, process_method):
            return getattr(self, process_method)(text, target, options)
        else:
            raise ValueError("Don't know how to render target '{}'".format(target))

    def escape(self, text):
        return text_type(text)

    def process_text(self, text, target, options):
        html_markup = self.process_html(text, target, options)
        return html.textilize(html_markup)

    def moya_render(self, archive, context, target, options):
        rendered = self.process(self.source, target, options)
        return rendered


class MarkupBase(with_metaclass(MarkupBaseMeta, MarkupBaseType)):
    pass


class TextMarkup(MarkupBase):
    name = "text"

    def process_html(self, text, target, options):
        return HTML(html.escape(text).replace('\n', '<br>'))


class HTMLMarkup(MarkupBase):
    name = "html"

    def process_html(self, text, target, options):
        return HTML(text)


class BBCodeMarkup(MarkupBase):
    name = "bbcode"

    def process_text(self, text, target, options):
        return postmarkup.strip_bbcode(text)

    def process_html(self, text, target, options):
        html = postmarkup.render_bbcode(text)
        return HTML(html)


# class MarkdownMarkup(MarkupBase):
#     name = "markdown"

#     def create(self, options):
#         if 'output_format' not in options:
#             options['output_format'] = 'html5'
#         self.markdown = markdown.Markdown(**options)

#     def escape(self, text):
#         return html.escape(text)

#     def process_html(self, text, target, options):
#         return HTML(self.markdown.convert(text))


class SummaryMarkup(MarkupBase):
    name = "summary"

    def process_html(self, text, target, options):
        return HTML(html.summarize(text, max_size=options.get('length', 100)))


class MarkdownMarkup(MarkupBase):
    name = "markdown"

    def create(self, options):
        self.parser = CommonMark.DocParser()
        self.renderer = CommonMark.HTMLRenderer()

    def process_html(self, text, target, options):
        ast = self.parser.parse(text)
        html = HTML(self.renderer.render(ast))
        return html


@implements_to_string
class Markup(object):
    html_safe = True

    def __init__(self, source, type, markup_options=None):
        self.source = source
        self.type = type
        if markup_options is None:
            markup_options = {}
        self.markup_options = markup_options
        try:
            self.markup_processor = MarkupBase.get_processor(self.type, markup_options)
        except KeyError:
            raise ValueError("No markup processor called '{}'".format(self.type))

    @classmethod
    def get_escape(cls, type, markup_options=None):
        try:
            markup_processor = MarkupBase.get_processor(type, markup_options)
        except KeyError:
            raise ValueError("No markup processor called '{}'".format(type))
        return markup_processor.escape

    @classmethod
    def sub(cls, type, context, text, markup_options=None):
        try:
            markup_processor = MarkupBase.get_processor(type, markup_options)
        except KeyError:
            raise ValueError("No markup processor called '{}'".format(type))
        return context.sub(text, markup_processor.escape)

    @classmethod
    def supports(cls, name):
        return name in MarkupBaseMeta.markup_registry

    def __str__(self):
        return self.markup_processor.process(self.source, "text", None)

    def moya_render(self, archive, context, target, options):
        return self.markup_processor.process(self.source, target, options)

    def process(self, target="html", **options):
        return self.markup_processor.process(self.source, target, options)



from __future__ import unicode_literals
from __future__ import print_function

from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...console import Cell


class Apps(SubCommand):
    """List project applications"""
    help = "get application information"

    def add_arguments(self, parser):
        parser.add_argument("-l", '--location', dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to projects settings file")
        return parser

    def run(self):
        application = WSGIApplication(self.location, self.get_settings(), disable_autoreload=True)
        archive = application.archive
        table = []

        for name, app in sorted(archive.apps.items()):
            table.append([
                         name,
                         app.lib.long_name,
                         Cell(app.lib.version, bold=True, fg="magenta"),
                         ])
        self.console.table(table,
                           ['app', 'lib', 'version'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys
import platform
import warnings
import re
from threading import RLock
from textwrap import wrap, dedent
warnings.filterwarnings("ignore")
from os.path import splitext

try:
    import colorama
except ImportError:
    colorama = None
else:
    colorama.init()


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3
WIN = sys.platform.startswith('win')

if PY3:
    text_type = str
    binary_type = bytes
    xrange = range
    string_types = (str, bytes)
else:
    text_type = unicode
    binary_type = str
    string_types = (basestring,)


class ConsoleHighlighter(object):

    styles = {}
    highlights = []

    @classmethod
    def highlight(cls, text, default_style=''):
        cls.compile()
        text = AttrText(text)
        default_style = style(default_style or cls.styles.get(None, ''))

        if default_style:
            text.add_span(0, len(text), **default_style)

        for highlight in cls._compiled_highlights:
            for match in highlight.finditer(text):
                for k, v in match.groupdict().items():
                    if k in cls._compiled_styles and v:
                        start, end = match.span(k)
                        text.add_span(start, end, **cls._compiled_styles[k])
        return text

    @classmethod
    def compile(cls):
        if not hasattr(cls, '_compiled_highlights'):
            cls._compiled_highlights = [re.compile(h, re.UNICODE | re.DOTALL | re.MULTILINE) for h in cls.highlights]
            cls._compiled_styles = {k: style(v) for k, v in cls.styles.items()}


class XMLHighlighter(ConsoleHighlighter):
    styles = {
        "xmlns": "italic",
        "tag": "bold blue not dim",
        "attribute": "not bold cyan",
        "string": "yellow",
        "substitution": "bold magenta",
        "braced": "bold",
        "comment": "dim white italic"
    }

    highlights = [
        r'(?P<tag>\<.*?\>)',
        r'(?P<attribute>\s\S*?=\".*?\")',
        r'(?P<string>\".*?\")',
        r'(?P<braced>\{.*?\})',
        r'(?P<substitution>\$\{.*?\})',
        r'(?P<comment>\<\!\-\-.*?\-\-\>)',
    ]


class TemplateHighlighter(ConsoleHighlighter):
    styles = {
        "tag": "bold blue not dim",
        "attr": "cyan not bold",
        "string": "yellow",
        "substitution": "bold magenta",
        "templatetag": "bold black",
        "comment": "dim white italic"
    }

    highlights = [
        r'(?P<tag>\<.*?\>)',
        r'(?P<attr>\s\S*?=\".*?\")',
        r'(?P<string>\".*?\")',
        r'(?P<substitution>\$\{.*?\})',
        r'(?P<templatetag>\{\%.*?\%\})',
        r'(?P<comment>\<\!\-\-.*?\-\-\>)',
    ]


def style(style_def):
    """Convert a console style definition in to dictionary

    >>> style("bold red on yellow")
    {fg="red", bg="yellow", bold=True}

    """
    if not style_def:
        return {}
    if isinstance(style_def, dict):
        return style_def
    colors = {"yellow", "magenta", "green", "cyan", "blue", "red", "black", "white"}
    text_styles = {"bold", "underline", "dim", "reverse", "italic"}
    style = {}
    foreground = True
    style_set = True
    for s in style_def.split(' '):
        if s == "on":
            foreground = False
        elif s == "not":
            style_set = False
        elif s in colors:
            style["fg" if foreground else "bg"] = s
        elif s in text_styles:
            style[s] = style_set
        else:
            raise ValueError("unknown style '{}'".format(s))
    return style


class AttrText(text_type):
    """A string with associate console attribute information"""
    def __init__(self, text, spans=None, *args, **kwargs):
        super(AttrText, self).__init__()
        self.attr_spans = spans or []

    def __repr__(self):
        return 'AttrText(%r)' % super(AttrText, self).__repr__()

    def add_span(self, start, end=None, **attrs):
        """Apply attributes to a span in the string"""
        #if end < 0 or start > len(self):
        #    return
        if end is None:
            end = len(self)
        self.attr_spans.append((max(0, start),
                                min(len(self), end),
                                attrs))

    def splitlines(self):
        """Split text in to lines, preserving attributes"""
        bucket_shift = 6
        lines = [[] for _ in xrange((len(self) >> bucket_shift) + 1)]
        pos = 0
        new_lines = []
        line_count = 0
        find = self.find
        l = len(self)
        while pos < l:
            line_end = find('\n', pos)
            if line_end == -1:
                line_end = len(self)  # - 1
            new_lines.append(AttrText(self[pos:line_end]))
            for line_no in xrange(pos >> bucket_shift, (line_end >> bucket_shift) + 1):
                lines[line_no].append((pos, line_end, line_count))
            line_count += 1
            pos = line_end + 1

        for start, end, attrs in self.attr_spans:
            for line_list in lines[start >> bucket_shift:(end >> bucket_shift) + 1]:
                for line_start, line_end, line_offset in line_list:
                    line = new_lines[line_offset]
                    line.attr_spans.append((max(0, start - line_start), min(len(line), end - line_start), attrs))

        return new_lines

    def __moyaconsole__(self, console):
        """Write to a console (called by Console.text)"""
        chars = list(self)
        attrs = [{} for c in chars]
        for start, end, span_attrs in self.attr_spans:
            for r in range(start, end):
                attrs[r].update(span_attrs)

        last_attrs = {}
        accum = []
        text = []
        accum_append = accum.append
        text_append = text.append

        for c, c_attrs in zip(chars, attrs):
            if c_attrs == last_attrs:
                accum_append(c)
            else:
                if accum:
                    span_text = ''.join(accum)
                    text_append((span_text, last_attrs))
                del accum[:]
                accum_append(c)
                last_attrs = c_attrs

        if accum:
            span_text = ''.join(accum)
            text_append((span_text, last_attrs))

        console_out = console.__call__

        with console._lock:
            for text, attrs in text:
                console_out(text, **attrs)

    # def highlight(self, re_attributes, compile=False, default=None):
    #     if compile:
    #         re_attributes = compile_highlights(re_attributes)
    #     if default is not None:
    #         self.add_span(0, len(self), **default)
    #     for re_text, attrs in re_attributes:
    #         for match in re_text.finditer(self):
    #             self.add_span(match.start(), match.end(), **attrs)
    #     return self


if platform.system() == 'Windows':

    def getTerminalSize():
        try:
            ## {{{ http://code.activestate.com/recipes/440694/ (r3)
            from ctypes import windll, create_string_buffer

            # stdin handle is -10
            # stdout handle is -11from
            # stderr handle is -12

            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)

            if res:
                import struct
                (bufx, bufy, curx, cury, wattr,
                 left, top, right, bottom, maxx, maxy) = struct.unpack(b"hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                sizey = bottom - top + 1
            else:
                sizex, sizey = 80, 25  # can't determine actual size - return default values
            return sizex, sizey
        except:
            return 80, 25

else:

    def getTerminalSize():
        def ioctl_GWINSZ(fd):
            try:
                import fcntl
                import termios
                import struct
                import os
                cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            except:
                return None
            return cr
        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            import os
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except:
                pass
        if cr:
            return int(cr[1]), int(cr[0])
        try:
            h, w = os.popen("stty size", "r").read().split()
            return int(w), int(h)
        except:
            pass
        return 80, 25


class Cell(object):
    """A table cell"""
    def __init__(self, text, processor=None, **attribs):
        self.text = text_type(text)
        self.processor = processor or (lambda t: t)
        self.attribs = attribs

    @classmethod
    def create(cls, cell, processor=None):
        if isinstance(cell, Cell):
            return cell
        if cell is None:
            return Cell('None', italic=True, dim=True)
        try:
            text = text_type(cell)
        except:
            text = repr(cell)
        attribs = {}
        return Cell(text, processor=processor, **attribs)

    def __len__(self):
        return 2

    def __iter__(self):
        yield self.processor(self.text)
        yield self.attribs

    def __getitem__(self, index):
        if index == 0:
            return self.processor(self.text)
        if index == 1:
            return self.attribs
        raise IndexError(index)

    def __repr__(self):
        return 'Cell(%r)' % self.text

    def get_min_length(self):
        try:
            return min(max(len(token) for token in self.text.split(splitter))
                       for splitter in (' ', ',', '/'))
        except ValueError:
            return 1

    def get_lines(self, max_length=80, continuation='↪'):
        center = self.attribs.get("center", False)
        if max_length is None:
            return self.text.splitlines()
        lines = []
        if max_length < 8:
            continuation = ''

        def add_line(line):
            if not line:
                lines.append('')
                return
            new_lines = []
            while len(line) > max_length:
                broken_line = line[:max_length]
                break_pos = max_length
                for break_char in ' /.-_':
                    if break_char in broken_line[8:]:
                        break_pos = broken_line.rindex(break_char)
                        break
                new_line = line[:break_pos]
                remaining_line = continuation + line[break_pos:]
                new_lines.append(new_line)
                line = remaining_line
            if line:
                new_lines.append(line)
            if center:
                new_lines = [l.center(max_length) for l in new_lines]
            lines.extend(new_lines)

        for line in self.text.splitlines():
            tokens = line.split(' ')[::-1]
            line_tokens = []

            while tokens:
                token = tokens.pop()
                line_tokens.append(token)
                new_line = " ".join(line_tokens)
                if len(new_line) > max_length:
                    if len(line_tokens) > 1:
                        line_tokens.pop()
                        add_line(" ".join(line_tokens))
                        del line_tokens[:]
                        tokens.append(token)
                    else:
                        add_line(new_line)
                        del line_tokens[:]

            if line_tokens:
                add_line(" ".join(line_tokens))

        return lines

    def expand(self, max_length=None, continuation='↪'):
        lines = self.get_lines(max_length=max_length, continuation=continuation)
        expanded_lines = [Cell(line, processor=self.processor, **self.attribs) for line in lines]
        return expanded_lines


def make_table_header(*headers):
    """Creates the first row of headers in a table"""
    return [[Cell(h, bold=True) for h in headers]]


class _TextOut(object):
    def __init__(self):
        self.text = []

    def write(self, text):
        if not isinstance(text, text_type):
            text = text_type(text, 'utf-8')
        self.text.append(text)

    def getvalue(self):
        return ''.join(self.text)


class Console(object):
    """Write output to the console, with styles and color"""

    fg_colors = dict(black=30,
                     red=31,
                     green=32,
                     yellow=33,
                     blue=34,
                     magenta=35,
                     cyan=36,
                     white=37)

    bg_colors = dict(black=40,
                     red=41,
                     green=42,
                     yellow=43,
                     blue=44,
                     magenta=45,
                     cyan=46,
                     white=47)

    _lock = RLock()

    def __init__(self, out=None, nocolors=False, text=False, width=None, html=False):
        #self._lock = RLock()
        if html:
            self.html = True
            self.encoding = 'utf-8'
            self.out = _TextOut()
            self.unicode_borders = True
            self.terminal_width = width or 120
            return
        if text:
            nocolors = True
            out = _TextOut()
        self.out = out or sys.stdout
        self.html = html

        self.encoding = getattr(self.out, 'encoding', 'utf-8') or 'utf-8'
        if nocolors:
            self.terminal_colors = False
        else:
            self.terminal_colors = self.is_terminal()
            if sys.platform.startswith('win') and not colorama:
                self.terminal_colors = False
        if self.is_terminal():
            w, h = getTerminalSize()
            self.terminal_width = w
        else:
            self.terminal_width = width or 80
        self.unicode_borders = self.terminal_colors and not WIN

    @property
    def width(self):
        return self.terminal_width

    def flush(self):
        with self._lock:
            if hasattr(self.out, 'flush'):
                self.out.flush()

    def update_terminal_width(self):
        with self._lock:
            if self.is_terminal():
                w, h = getTerminalSize()
                self.terminal_width = w

    def get_text(self):
        assert isinstance(self.out, _TextOut)
        return self.out.getvalue()

    def __repr__(self):
        return "<console>"

    def is_terminal(self):
        try:
            return self.out.isatty()
        except AttributeError:
            return False

    def _html_out(self,
                  text,
                  fg=None,
                  bg=None,
                  bold=False,
                  underline=False,
                  dim=False,
                  reverse=False,
                  italic=False,
                  nl=False):
        css_classes = []
        append = css_classes.append
        if bold:
            append('console-bold')
        if underline:
            append('console-underline')
        if dim:
            append('console-dim')
        if reverse:
            append('console-reverse')
        if italic:
            append('console-italic')
        if fg is not None:
            append('console-foreground-' + fg)
        if bg is not None:
            append('console-background-' + bg)
        if italic:
            append('console-italic')

        class_attrib = " ".join(css_classes)

        if class_attrib:
            tag = '<span class="{}">'.format(class_attrib)
            self.out.write(tag)

        text = text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
        text = text.replace('\n', '<br>')
        self.out.write(text.encode('utf-8'))

        if class_attrib:
            self.out.write('</span>')

        if nl:
            self.out.write('<br>')

        return self

    def __call__(self,
                 text,
                 fg=None,
                 bg=None,
                 bold=False,
                 underline=False,
                 dim=False,
                 reverse=False,
                 italic=False,
                 nl=False,
                 asstr=False,
                 center=False):
        if isinstance(text, AttrText):
            return self.text(text)
        #if not isinstance(text, string_types):
        #    self.obj(context, text)
        #    return self
        if self.html:
            return self._html_out(text,
                                  fg=fg,
                                  bg=bg,
                                  bold=bold,
                                  underline=underline,
                                  dim=dim,
                                  reverse=reverse,
                                  italic=italic,
                                  nl=nl)
        if not isinstance(text, text_type):
            text = text.decode('ascii', 'replace')
        if PY2:
            text = text.encode(self.encoding, 'replace')

        if nl:
            text += '\n'
        if not self.terminal_colors:
            if asstr:
                return text
            self.out.write(text)
            return self
        out = []
        fg = self.fg_colors.get(fg, None) if fg is not None else None
        bg = self.bg_colors.get(bg, None) if bg is not None else None
        attrs = []
        if bold:
            attrs.append(1)
        if dim:
            attrs.append(2)
        if italic:
            attrs.append(3)
        if underline:
            attrs.append(4)
        if reverse:
            attrs.append(7)

        attrs.append(fg)
        attrs.append(bg)

        display_attributes = ["%i" % da for da in attrs if da is not None]

        if display_attributes:
            out.append('\x1b[%sm' % ';'.join(display_attributes))
        out.append(text)
        if display_attributes:
            out.append("\x1b[0m")
        if asstr:
            return ''.join(out)

        if PY3:
            def console_encode(s):
                """Work around a bug with colorama on Windows"""
                if self.encoding.lower() != 'utf-8':
                    return s.encode(self.encoding, 'replace').decode(self.encoding)
                return s
            with self._lock:
                self.out.write(''.join(
                               (text.decode('utf-8', 'replace') if isinstance(text, bytes) else console_encode(text))
                               for text in out))
        else:
            with self._lock:
                self.out.write(b''.join(
                               (text.encode(self.encoding, 'replace') if isinstance(text, text_type) else text)
                               for text in out))
        return self

    def progress(self, msg, num_steps=100, width=12):
        """A context manager to manage progress bars"""
        from .progress import Progress, ProgressContext
        p = Progress(self, msg, num_steps=num_steps, width=width)
        p.render()
        return ProgressContext(p)

    def nl(self, count=1):
        with self._lock:
            if self.html:
                self.out.write('<br>' * count)
            else:
                self.out.write('\n' * count)
        return self

    def div(self, msg=None, **attrs):
        """Inserts a horizontal dividing line"""
        if 'italic' not in attrs:
            attrs['italic'] = True
        with self._lock:
            self.update_terminal_width()
            if msg is None:
                self("-" * self.terminal_width, dim=True).nl()
            else:
                space = self.terminal_width - len(msg)
                lspace = space // 2
                rspace = space - lspace
                self('-' * lspace, dim=True)(msg, **attrs)('-' * rspace, dim=True).nl()

        return self

    def text(self, text, **params):
        with self._lock:
            if isinstance(text, AttrText):
                text.__moyaconsole__(self)
            else:
                self(text, **params).nl()
        return self

    def wraptext(self, text, do_dedent=True, **attribs):
        """Output wrapper text"""
        with self._lock:
            if do_dedent:
                text = dedent(text)
            for line in text.splitlines():
                wrapped_text = wrap(line, self.terminal_width)
                self.text("\n".join(wrapped_text).lstrip(), **attribs)
        return self

    def xmlsnippet(self, code, lineno=1, colno=None, extralines=3, line_numbers=True):
        """Render a snippet of xml, with a highlighted line"""
        with self._lock:
            if not code:
                return
            if colno is not None:
                highlight_columns = (colno - 1, colno)
            else:
                highlight_columns = None
            self.snippet(code,
                         (lineno - extralines, lineno + extralines + 1),
                         highlight_line=lineno,
                         highlight_columns=highlight_columns,
                         line_numbers=True)
        return self

    def templatesnippet(self, code, lineno=1, colno=None, endcolno=None, extralines=3, line_numbers=True):
        with self._lock:
            if not code:
                return
            if colno is not None:
                highlight_columns = (colno - 1, colno if endcolno is None else endcolno)
            else:
                highlight_columns = None
            self.snippet(code,
                         (lineno - extralines, lineno + extralines + 1),
                         highlight_line=lineno,
                         highlight_columns=highlight_columns,
                         highlighter=TemplateHighlighter,
                         line_numbers=line_numbers)
        return self

    def xml(self, code):
        code = code.strip('\n') + '\n'
        self.snippet(code, line_numbers=False)
        return self
        with self._lock:
            self.update_terminal_width()
            if not self.terminal_colors:
                self(code).nl()
            else:
                try:
                    from pygments import highlight
                    from pygments.lexers import XmlLexer
                    from pygments.formatters import TerminalFormatter
                except ImportError:
                    self(xml).nl()
                else:
                    hcode = highlight(code, XmlLexer(), TerminalFormatter())
                    self(hcode)
            return self

    def document_error(self, msg, path, code, lineno, colno, diagnosis=None):
        with self._lock:
            self.div()
            if colno is None:
                self('File "%s", line %i' % (path, lineno)).nl()
            else:
                self('File "%s", line %i, column %i' % (path, lineno, colno)).nl()
            self(msg, fg="red", bold=True).nl()
            if diagnosis:
                self.table([[Cell(diagnosis, italic=True)]])
            self.xmlsnippet(code, lineno, colno)

        return self

    def success(self, msg):
        with self._lock:
            self.wraptext(msg, fg="green", bold=True)
        return self

    def error(self, msg):
        """Renders a generic error"""
        with self._lock:
            self.wraptext(msg, fg="red", bold=True)
        return self

    def exception(self, exc, tb=False):

        with self._lock:
            self.update_terminal_width()

            if not tb:
                if hasattr(exc, 'get_moya_error'):
                    exc_text = exc.get_moya_error()
                else:
                    exc_text = text_type(exc)
                self(exc_text, fg="red", bold=True).nl()
                return self

            if isinstance(exc, string_types):
                raw_tb = exc
            else:
                import traceback
                raw_tb = traceback.format_exc()

            if self.terminal_colors:
                from pygments import highlight
                from pygments.lexers import PythonTracebackLexer, Python3TracebackLexer
                from pygments.formatters import TerminalFormatter

                if PY2:
                    lexer = PythonTracebackLexer
                else:
                    lexer = Python3TracebackLexer

                htb = highlight(raw_tb,
                                lexer(),
                                TerminalFormatter())
                self(htb)
            else:
                self(raw_tb)

        return self

    def table(self,
              table,
              header_row=None,
              grid=True,
              header=True,
              divider_attribs=None,
              pad=1,
              border_style=0,
              dividers=True,
              cell_processors=None):
        """Renders a table of cells with an optional ASCII grid

        A table should be a list of lists, where each element is either a string
        or a tuple of a string and a dictionary of attributes.

        """
        table = list(table)
        with self._lock:
            if cell_processors is None:
                cell_processors = {}
            if header_row is not None and header:
                table = make_table_header(*header_row) + table[:]
            tl = tr = bl = br = ir = it = il = ib = ii = '+'
            hor = '-'
            ver = '|'
            continuation = '...'
            if self.unicode_borders:
                continuation = '↪  '
                if border_style == 0:
                    tl = '╭'
                    tr = '╮'
                    bl = '╰'
                    br = '╯'
                    il = '├'
                    ir = '┤'
                    it = '┬'
                    ib = '┴'
                    hor = '─'
                    ver = '│'
                    ii = '┼'
                elif border_style == 1:
                    tl = '╔'
                    tr = '╗'
                    bl = '╚'
                    br = '╝'
                    il = '╠'
                    ir = '╣'
                    it = '╦'
                    ib = '╩'
                    hor = '═'
                    ver = '║'
                    ii = '╬'
                elif border_style == 2:
                    tl = '┏'
                    tr = '┓'
                    bl = '┗'
                    br = '┛'
                    il = '┣'
                    ir = '┫'
                    it = '┳'
                    ib = '┻'
                    hor = '━'
                    ver = '┃'
                    ii = '╋'

            self.update_terminal_width()
            terminal_width = self.terminal_width
            if WIN:
                terminal_width -= 1
            table = [[Cell.create(cell, cell_processors.get(rowno))
                      for rowno, cell in enumerate(row)] for row in table]

            def cell_len(cell):
                try:
                    return max(len(line) for line in cell.get_lines(max_length=terminal_width))
                except ValueError:
                    return 0

            cell_lengths = []
            cell_min_lengths = []
            for row_no in xrange(len(table[0])):
                cell_lengths.append(max(cell_len(col[row_no]) for col in table))
                cell_min_lengths.append(max(col[row_no].get_min_length() for col in table))

            num_cols = len(cell_lengths)
            table_padding = (num_cols * pad * 2) + (num_cols - 1) + grid * 2
            table_width = sum(cell_lengths) + table_padding
            if table_width > self.terminal_width:

                # make each column its minimum until the table fits, starting with the widest column
                for i, (cell_length, min_length) in sorted(enumerate(zip(cell_lengths, cell_min_lengths)), key=lambda c: c[1][1], reverse=True):
                    over_size = table_width - terminal_width
                    cell_lengths[i] -= min(over_size, (cell_length - min_length))
                    table_width = sum(cell_lengths) + table_padding
                    if sum(cell_lengths) <= terminal_width:
                        break

                over_space = table_width - terminal_width
                while over_space > 0:
                    largest_value = 0
                    largest_index = None
                    for i, l in enumerate(cell_lengths):
                        if l > largest_value:
                            largest_value = l
                            largest_index = i
                    if largest_index is None:
                        break
                    reduce = min(over_space, int(largest_value - 4))
                    cell_lengths[largest_index] -= reduce
                    if reduce <= 0:
                        break
                    over_space -= reduce

            if grid:
                if divider_attribs is None:
                    divider_attribs = {"dim": True}
                top_divider = (tl + '%s' + tr) % it.join(hor * (l + pad * 2) for l in cell_lengths)
                mid_divider = (il + '%s' + ir) % ii.join(hor * (l + pad * 2) for l in cell_lengths)
                bot_divider = (bl + '%s' + br) % ib.join(hor * (l + pad * 2) for l in cell_lengths)
            else:
                divider_attribs = {}
                divider = ''
                top_divider = ''
                mid_divider = ''
                bot_divider = ''

            if grid:
                self(top_divider, **divider_attribs).nl()

            padding = pad * ' '
            separator = ver if grid else ' '
            for row_no, row in enumerate(table):
                if row_no == len(table) - 1:
                    divider = bot_divider
                else:
                    divider = mid_divider
                expanded_cells = [cell.expand(cell_length, continuation=continuation)
                                  for cell_length, cell in zip(cell_lengths, row)]
                max_height = max(len(cell) for cell in expanded_cells)

                for cell_line_no in xrange(max_height):
                    expanded_row = [ecell[cell_line_no] if cell_line_no < len(ecell) else Cell('')
                                    for ecell in expanded_cells]
                    if grid:
                        r = [(ver, divider_attribs)]
                    else:
                        r = []
                    append = r.append

                    last_i = len(expanded_row) - 1
                    for i, (cell, cell_length) in enumerate(zip(expanded_row, cell_lengths)):
                        text, attribs = cell
                        if grid:
                            cell_text = (padding + text.ljust(cell_length) + padding)
                        else:
                            cell_text = (padding * (i > 0) + text.ljust(cell_length) + padding * (i < last_i))
                        cell_text = cell.processor(cell_text)
                        append((cell_text, attribs))
                        if grid or i < last_i:
                            append((separator, divider_attribs))

                    for text, attribs in r:
                        self(text, **attribs)
                    self.nl()

                if grid:
                    if dividers and not header:
                        self(divider, **divider_attribs).nl()
                    elif not dividers:
                        if row_no == len(table) - 1:
                            self(divider, **divider_attribs).nl()
                    elif row_no in (0, len(table) - 1):
                        self(divider, **divider_attribs).nl()
            return self

    def cat(self, contents, path):
        with self._lock:
            self.update_terminal_width()
            if not self.terminal_colors:
                self(contents).nl()
            else:

                ext = splitext(path)[-1].lower()
                if ext in ('.htm', '.html', '.text'):
                    self(TemplateHighlighter.highlight(contents))
                    return self
                elif ext in ('.xml',):
                    self(XMLHighlighter.highlight(contents))
                    return self

                from pygments import highlight
                from pygments.formatters import TerminalFormatter
                from pygments.lexers import get_lexer_for_filename, guess_lexer
                from pygments.util import ClassNotFound
                try:
                    lexer = guess_lexer(contents)
                except ClassNotFound:
                    try:
                        lexer = get_lexer_for_filename(path)
                    except ClassNotFound:
                        lexer = None

                if lexer is None:
                    self(contents).nl()
                else:
                    hcode = highlight(contents, lexer, TerminalFormatter())
                    self(hcode)
            return self

    def snippet(self, xml, line_range=None, line_numbers=True, highlight_line=None, highlight_columns=None, highlighter=None):
        if isinstance(xml, binary_type):
            xml = xml.decode('utf-8', 'replace')
        with self._lock:
            self.update_terminal_width()
            if line_range is None:
                start = 1
                end = None
            else:
                start, end = line_range

            #xml = xml.replace('\r', '\n')
            xml = AttrText(xml)

            if highlighter is None:
                highlighter = XMLHighlighter

            xml = highlighter.highlight(xml)

            lines = xml.splitlines()

            if end is None:
                end = len(lines) + 1
            if start < 1:
                start = 1
            if end > len(lines):
                end = len(lines) + 1
            lines = lines[start - 1:end - 1]

            try:
                max_number_length = max(len(text_type(n + 1)) for n in range(start, end)) + 1
            except:
                max_number_length = 0
            if max_number_length < 6:
                max_number_length = 6

            if line_numbers:
                for i, line in enumerate(lines):
                    line_no = i + start
                    if highlight_line is not None and highlight_line == line_no:
                        number = ("*" + text_type(line_no)).rjust(max_number_length, ' ') + ' '

                        if highlight_columns:
                            col_start, col_end = highlight_columns
                            line.add_span(col_start, col_end, bold=True, fg="white", bg="red")

                        self(number, bold=True)(line).nl()
                    else:
                        number = text_type(line_no).rjust(max_number_length) + ' '
                        self(number, bold=True, fg="black")(line).nl()
            else:
                for line in lines:
                    self(line).nl()
        return self

    def obj(self, context, obj, **kwargs):
        """Writes information regarding an object to the console"""
        with self._lock:
            if hasattr(obj, '__moyaconsole__'):
                try:
                    obj.__moyaconsole__(self)
                except Exception:
                    pass
                else:
                    return self

            if isinstance(obj, string_types):
                self.text(context.to_expr(obj), **kwargs)
            elif isinstance(obj, dict):
                table = []
                for k, v in obj.items():
                    table.append([k, context.to_expr(v)])
                self.table(table, header_row=["key", "value"])
            elif isinstance(obj, (list, tuple)):
                table = [(i, context.to_expr(v)) for i, v in enumerate(obj)]
                self.table(table, header_row=["index", "value"])
            elif isinstance(obj, bool):
                if obj:
                    self.text("True", bold=True, fg="green")
                else:
                    self.text("False", bold=True, fg="red")
            else:
                self.text(context.to_expr(obj), **kwargs)
            return self

    def show_cursor(self, show=True):
        if not WIN:
            if show:
                self.out.write("\033[?25h")
            else:
                self.out.write("\033[?25l")
        self.out.flush()


def test_table(console):
    long_text = """If you are writing an application of any size, it will most likely require a number of files to run - files which could be stored in a variety of possible locations. Furthermore, you will probably want to be able to change the location of those files when debugging and testing. You may even want to store those files somewhere other than the user's hard drive."""

    bold = dict(bold=True)
    red = dict(fg="red")

    table = [
             [("foo", dict(reverse=True, bold=True, dim=True)), ("100", red), "Some text"],
             [("long text!", dict(reverse=True, bold=True, dim=True)), long_text, "Some text"],
             [("bar", bold), ("120", red), ("Some more text\nline 2\nline 3", {'bg':'green'})],
             [("foo", bold), ("100", red), "Some text"],
             [("bar", bold), ("120", red), "Some more text"],
             [("A longer cell...\nwith multiple lines", bold), (120344, red), "Some more text"],
             [("foo", bold), ("100", red), "Some text"],
             [("bar", bold), ("120", red), "Some more text"],
            ]
    console.table(table, grid=True)
    console.nl()


if __name__ == "__main__":

    console = Console()
    test_table(console)

    import sys
    #sys.exit()

    xml = """<moya xmlns="http://moyaproject.com">
    <mountpoint name="main">
        <url url="{*path}">
            <call py="static.check_hide" dst="hidden" >
                <set value="url.path" />
            </call>
            <if test="url.path $= '/'">
                <call py="static.get_dirlist" dst="dirlist">
                    <set value=".fs[app.settings.fs]" />
                    <set value="url.path" />
                </call>
                <servepage template="dirlist.html" withscope="y" if="dirlist" />
            </if>
            <else>
                <serve fs="${app.settings.fs}" path="${url.path}" />
            </else>
        </url>
    </mountpoint>
</moya>
<!-- Commented <b>out</b> -->
"""
    console = Console()
    #console.xml_trace(xml, (4, 11), highlight_line=8)
    console.snippet(xml, (4, 11), highlight_line=7, highlight_columns=(12,100))

if 0:
    console = Console()
    console('Hello ', dim=True)("World!", bold=True).nl()
    console("Green!", bg="green", fg="white").nl()

    xml = """<moya xmlns="http://moyaproject.com">

<mountpoint name="testmount" libname="root">
    <url name="article" url="/{year}/{month}/{day}/{slug}/" methods="GET" target="viewpost">
        <debug>url main: ${url.year}, ${url.month}, ${url.day}, ${url.slug}</debug>
    </url>
    <url name="front" url="/" methods="GET">
        <debug>Front...</debug>
        <return><str>Front</str></return>
    </url>
</mountpoint>

<macro docname="viewpost">
    <debugIn viewpost</debug>
    <return><str>Hello, World</str></return>
    <return>
        <response template="birthday.html">
            <str dst="title">My Birthday</str>
            <str dst="body">It was my birthday today!</str>
        </response>
    </return>
</macro>

<!--
<macro libname="showapp">
    <debug>App is ${app}</debug>
</macro>

<macro libname="blogmacro">
    <debug>Called blogmacro in blog lib</debug>
</macro>

<macro libname="blogmacro2">
    <debug>Called blogmacro2 with app: ${debug:app}</debug>
</macro>
-->

</moya>"""*10

    console.xmlsnippet(xml, 6)

    try:
        1 / 0
    except Exception as e:
        console.exception(e, tb=True)

    bold = dict(bold=True)
    red = dict(fg="red")

    long_text = """If you are writing an application of any size, it will most likely require a number of files to run - files which could be stored in a variety of possible locations. Furthermore, you will probably want to be able to change the location of those files when debugging and testing. You may even want to store those files somewhere other than the user's hard drive."""

    table = [
             [("foo", dict(reverse=True, bold=True, dim=True)), ("100", red), "Some text"],
             [("long text!", dict(reverse=True, bold=True, dim=True)), long_text, "Some text"],
             [("bar", bold), ("120", red), ("Some more text\nline 2\nline 3", {'bg':'green'})],
             [("foo", bold), ("100", red), "Some text"],
             [("bar", bold), ("120", red), "Some more text"],
             [("A longer cell...\nwith multiple lines", bold), (120344, red), "Some more text"],
             [("foo", bold), ("100", red), "Some text"],
             [("bar", bold), ("120", red), "Some more text"],
            ]
    console.table(table, grid=True)
    console.nl()
    console.table(table, grid=True, header=True)
    console.nl()
    console.table(table, grid=False, pad=0)

    #for line in Cell("""If you are writing an application of any size, it will most likely require a number of files to run - files which could be stored in a variety of possible locations. Furthermore, you will probably want to be able to change the location of those files when debugging and testing. You may even want to store those files somewhere other than the user's hard drive.""").get_lines():
    #        print line
